# This file is part of Supysonic.
# Supysonic is a Python implementation of the Subsonic server API.
#
# Copyright (C) 2013-2022 Alban 'spl0k' Féron
#                    2017 Óscar García Amor
#
# Distributed under terms of the GNU AGPLv3 license.

import unittest
import uuid

from supysonic.db import User

from .frontendtestbase import FrontendTestBase


class LoginTestCase(FrontendTestBase):
    def test_unauthorized_request(self):
        # Unauthorized request
        rv = self.client.get("/", follow_redirects=True)
        self.assertIn("Please login", rv.data)

    def test_login_with_bad_data(self):
        # Login with not blank user or password
        rv = self._login("", "")
        self.assertIn("Missing user name", rv.data)
        self.assertIn("Missing password", rv.data)
        # Login with not valid user or password
        rv = self._login("nonexistent", "nonexistent")
        self.assertIn("Wrong username or password", rv.data)
        rv = self._login("alice", "badpassword")
        self.assertIn("Wrong username or password", rv.data)

    def test_login_admin(self):
        # Login with a valid admin user
        rv = self._login("alice", "Alic3")
        self.assertIn("Logged in", rv.data)
        self.assertIn("Users", rv.data)
        self.assertIn("Folders", rv.data)

    def test_login_non_admin(self):
        # Login with a valid non-admin user
        rv = self._login("bob", "B0b")
        self.assertIn("Logged in", rv.data)
        # Non-admin user cannot acces to users and folders
        self.assertNotIn("Users", rv.data)
        self.assertNotIn("Folders", rv.data)

    def test_root_with_valid_session(self):
        # Root with valid session
        with self.client.session_transaction() as sess:
            sess["userid"] = User.get(name="alice").id
        rv = self.client.get("/", follow_redirects=True)
        self.assertIn("alice", rv.data)
        self.assertIn("Log out", rv.data)
        self.assertIn("There's nothing much to see here.", rv.data)

    def test_root_with_non_valid_session(self):
        # Root with a no-valid session
        with self.client.session_transaction() as sess:
            sess["userid"] = uuid.uuid4()
        rv = self.client.get("/", follow_redirects=True)
        self.assertIn("Please login", rv.data)

    def test_multiple_login(self):
        self._login("alice", "Alic3")
        rv = self._login("bob", "B0b")
        self.assertIn("Already logged in", rv.data)
        self.assertIn("alice", rv.data)


if __name__ == "__main__":
    unittest.main()



# This file is part of Supysonic.
# Supysonic is a Python implementation of the Subsonic server API.
#
# Copyright (C) 2017-2022 Alban 'spl0k' Féron
#
# Distributed under terms of the GNU AGPLv3 license.

import unittest
import uuid

from supysonic.db import Folder, Artist, Album, Track, User, ClientPrefs

from .apitestbase import ApiTestBase


class AnnotationTestCase(ApiTestBase):
    def setUp(self):
        super().setUp()

        root = Folder.create(name="Root", root=True, path="tests")
        folder = Folder.create(
            name="Folder", root=False, path="tests/assets", parent=root
        )
        artist = Artist.create(name="Artist")
        album = Album.create(name="Album", artist=artist)

        # Populate folder ids
        root = Folder.get(name="Root")
        folder = Folder.get(name="Folder")

        track = Track.create(
            title="Track",
            album=album,
            artist=artist,
            disc=1,
            number=1,
            path="tests/assets/empty",
            folder=folder,
            root_folder=root,
            duration=2,
            bitrate=320,
            last_modification=0,
        )

        self.folderid = folder.id
        self.artistid = artist.id
        self.albumid = album.id
        self.trackid = track.id
        self.user = User.get(name="alice")
        self.prefs = ClientPrefs.create(user=self.user, client_name="tests")

    def test_star(self):
        self._make_request("star", error=10)
        self._make_request("star", {"id": "unknown"}, error=0)
        self._make_request("star", {"albumId": "unknown"}, error=0)
        self._make_request("star", {"artistId": "unknown"}, error=0)
        self._make_request("star", {"id": str(uuid.uuid4())}, error=70)
        self._make_request("star", {"albumId": str(uuid.uuid4())}, error=70)
        self._make_request("star", {"artistId": str(uuid.uuid4())}, error=70)

        self._make_request("star", {"id": str(self.artistid)}, error=70)
        self._make_request("star", {"id": str(self.albumid)}, error=70)
        self._make_request("star", {"id": str(self.trackid)}, skip_post=True)
        self.assertIn(
            "starred", Track[self.trackid].as_subsonic_child(self.user, self.prefs)
        )
        self._make_request("star", {"id": str(self.trackid)}, error=0)

        self._make_request("star", {"id": str(self.folderid)}, skip_post=True)
        self.assertIn("starred", Folder[self.folderid].as_subsonic_child(self.user))
        self._make_request("star", {"id": str(self.folderid)}, error=0)

        self._make_request("star", {"albumId": str(self.folderid)}, error=0)
        self._make_request("star", {"albumId": str(self.artistid)}, error=70)
        self._make_request("star", {"albumId": str(self.trackid)}, error=70)
        self._make_request("star", {"albumId": str(self.albumid)}, skip_post=True)
        self.assertIn("starred", Album[self.albumid].as_subsonic_album(self.user))
        self._make_request("star", {"albumId": str(self.albumid)}, error=0)

        self._make_request("star", {"artistId": str(self.folderid)}, error=0)
        self._make_request("star", {"artistId": str(self.albumid)}, error=70)
        self._make_request("star", {"artistId": str(self.trackid)}, error=70)
        self._make_request("star", {"artistId": str(self.artistid)}, skip_post=True)
        self.assertIn("starred", Artist[self.artistid].as_subsonic_artist(self.user))
        self._make_request("star", {"artistId": str(self.artistid)}, error=0)

    def test_unstar(self):
        self._make_request(
            "star",
            {
                "id": [str(self.folderid), str(self.trackid)],
                "artistId": str(self.artistid),
                "albumId": str(self.albumid),
            },
            skip_post=True,
        )

        self._make_request("unstar", error=10)
        self._make_request("unstar", {"id": "unknown"}, error=0)
        self._make_request("unstar", {"albumId": "unknown"}, error=0)
        self._make_request("unstar", {"artistId": "unknown"}, error=0)

        self._make_request("unstar", {"id": str(self.trackid)}, skip_post=True)
        self.assertNotIn(
            "starred", Track[self.trackid].as_subsonic_child(self.user, self.prefs)
        )

        self._make_request("unstar", {"id": str(self.folderid)}, skip_post=True)
        self.assertNotIn("starred", Folder[self.folderid].as_subsonic_child(self.user))

        self._make_request("unstar", {"albumId": str(self.albumid)}, skip_post=True)
        self.assertNotIn("starred", Album[self.albumid].as_subsonic_album(self.user))

        self._make_request("unstar", {"artistId": str(self.artistid)}, skip_post=True)
        self.assertNotIn("starred", Artist[self.artistid].as_subsonic_artist(self.user))

    def test_set_rating(self):
        self._make_request("setRating", error=10)
        self._make_request("setRating", {"id": str(self.trackid)}, error=10)
        self._make_request("setRating", {"rating": 3}, error=10)
        self._make_request("setRating", {"id": "string", "rating": 3}, error=0)
        self._make_request(
            "setRating", {"id": str(uuid.uuid4()), "rating": 3}, error=70
        )
        self._make_request(
            "setRating", {"id": str(self.artistid), "rating": 3}, error=70
        )
        self._make_request(
            "setRating", {"id": str(self.albumid), "rating": 3}, error=70
        )
        self._make_request(
            "setRating", {"id": str(self.trackid), "rating": "string"}, error=0
        )
        self._make_request(
            "setRating", {"id": str(self.trackid), "rating": -1}, error=0
        )
        self._make_request("setRating", {"id": str(self.trackid), "rating": 6}, error=0)

        self.assertNotIn(
            "userRating", Track[self.trackid].as_subsonic_child(self.user, self.prefs)
        )

        for i in range(1, 6):
            self._make_request(
                "setRating", {"id": str(self.trackid), "rating": i}, skip_post=True
            )
            self.assertEqual(
                Track[self.trackid].as_subsonic_child(self.user, self.prefs)[
                    "userRating"
                ],
                i,
            )

        self._make_request(
            "setRating", {"id": str(self.trackid), "rating": 0}, skip_post=True
        )
        self.assertNotIn(
            "userRating", Track[self.trackid].as_subsonic_child(self.user, self.prefs)
        )

        self.assertNotIn(
            "userRating", Folder[self.folderid].as_subsonic_child(self.user)
        )
        for i in range(1, 6):
            self._make_request(
                "setRating", {"id": str(self.folderid), "rating": i}, skip_post=True
            )
            self.assertEqual(
                Folder[self.folderid].as_subsonic_child(self.user)["userRating"], i
            )
        self._make_request(
            "setRating", {"id": str(self.folderid), "rating": 0}, skip_post=True
        )
        self.assertNotIn(
            "userRating", Folder[self.folderid].as_subsonic_child(self.user)
        )

    def test_scrobble(self):
        self._make_request("scrobble", error=10)
        self._make_request("scrobble", {"id": "song"}, error=0)
        self._make_request("scrobble", {"id": str(uuid.uuid4())}, error=70)
        self._make_request("scrobble", {"id": str(self.folderid)}, error=0)

        self._make_request("scrobble", {"id": str(self.trackid)})
        self._make_request("scrobble", {"id": str(self.trackid), "submission": True})
        self._make_request("scrobble", {"id": str(self.trackid), "submission": False})


if __name__ == "__main__":
    unittest.main()



# This file is part of Supysonic.
# Supysonic is a Python implementation of the Subsonic server API.
#
# Copyright (C) 2018 Alban 'spl0k' Féron
#
# Distributed under terms of the GNU AGPLv3 license.

from flask import json, jsonify, make_response
from xml.etree import ElementTree

from . import API_VERSION


class BaseFormatter:
    def make_response(self, elem, data):
        raise NotImplementedError()

    def make_error(self, code, message):
        return self.make_response("error", {"code": code, "message": message})

    def make_empty(self):
        return self.make_response(None, None)

    def __call__(self, *args, **kwargs):
        return self.make_response(*args, **kwargs)

    error = make_error
    empty = property(make_empty)


class JSONBaseFormatter(BaseFormatter):
    def __remove_empty_lists(self, d):
        if not isinstance(d, dict):
            raise TypeError("Expecting a dict got " + type(d).__name__)

        keys_to_remove = []
        for key, value in d.items():
            if isinstance(value, dict):
                d[key] = self.__remove_empty_lists(value)
            elif isinstance(value, list):
                if len(value) == 0:
                    keys_to_remove.append(key)
                else:
                    d[key] = [
                        (
                            self.__remove_empty_lists(item)
                            if isinstance(item, dict)
                            else item
                        )
                        for item in value
                    ]

        for key in keys_to_remove:
            del d[key]

        return d

    def _subsonicify(self, elem, data):
        if (elem is None) != (data is None):
            raise ValueError("Expecting both elem and data or neither of them")

        rv = {"status": "failed" if elem == "error" else "ok", "version": API_VERSION}
        if data:
            rv[elem] = self.__remove_empty_lists(data)

        return {"subsonic-response": rv}


class JSONFormatter(JSONBaseFormatter):
    def make_response(self, elem, data):
        rv = jsonify(self._subsonicify(elem, data))
        rv.headers.add("Access-Control-Allow-Origin", "*")
        return rv


class JSONPFormatter(JSONBaseFormatter):
    def __init__(self, callback):
        self.__callback = callback

    def make_response(self, elem, data):
        if not self.__callback:
            return jsonify(
                self._subsonicify("error", {"code": 10, "message": "Missing callback"})
            )

        rv = self._subsonicify(elem, data)
        rv = f"{self.__callback}({json.dumps(rv)})"
        rv = make_response(rv)
        rv.mimetype = "application/javascript"
        return rv


class XMLFormatter(BaseFormatter):
    def __dict2xml(self, elem, dictionary):
        """Convert a dict structure to xml. The game is trivial. Nesting uses the [] parenthesis.
        ex.  { 'musicFolder': {'id': 1234, 'name': "sss" } }
          ex. { 'musicFolder': [{'id': 1234, 'name': "sss" }, {'id': 456, 'name': "aaa" }]}
          ex. { 'musicFolders': {'musicFolder' : [{'id': 1234, 'name': "sss" }, {'id': 456, 'name': "aaa" }] } }
          ex. { 'index': [{'name': 'A',  'artist': [{'id': '517674445', 'name': 'Antonello Venditti'}] }] }
          ex. {"subsonic-response": { "musicFolders": {"musicFolder": [{ "id": 0,"name": "Music"}]},
          "status": "ok","version": "1.7.0","xmlns": "http://subsonic.org/restapi"}}
        """
        if not isinstance(dictionary, dict):
            raise TypeError("Expecting a dict")
        if not all(isinstance(x, str) for x in dictionary):
            raise TypeError("Dictionary keys must be strings")

        for name, value in dictionary.items():
            if name == "value":
                elem.text = self.__value_tostring(value)
            elif isinstance(value, dict):
                subelem = ElementTree.SubElement(elem, name)
                self.__dict2xml(subelem, value)
            elif isinstance(value, list):
                for v in value:
                    subelem = ElementTree.SubElement(elem, name)
                    if isinstance(v, dict):
                        self.__dict2xml(subelem, v)
                    else:
                        subelem.text = self.__value_tostring(v)
            else:
                elem.set(name, self.__value_tostring(value))

    def __value_tostring(self, value):
        if value is None:
            return None
        if isinstance(value, str):
            return value
        if isinstance(value, bool):
            return str(value).lower()
        return str(value)

    def make_response(self, elem, data):
        if (elem is None) != (data is None):
            raise ValueError("Expecting both elem and data or neither of them")

        response = {
            "status": "failed" if elem == "error" else "ok",
            "version": API_VERSION,
            "xmlns": "http://subsonic.org/restapi",
        }
        if elem:
            response[elem] = data

        root = ElementTree.Element("subsonic-response")
        self.__dict2xml(root, response)

        rv = make_response(ElementTree.tostring(root))
        rv.mimetype = "text/xml"
        return rv



# This file is part of Supysonic.
# Supysonic is a Python implementation of the Subsonic server API.
#
# Copyright (C) 2013-2023 Alban 'spl0k' Féron
#
# Distributed under terms of the GNU AGPLv3 license.

import logging
import os
import os.path
import mediafile
import time

from datetime import datetime
from queue import Queue, Empty as QueueEmpty
from threading import Thread, Event

from .covers import find_cover_in_folder, CoverFile
from .db import Folder, Artist, Album, Track, open_connection, close_connection


logger = logging.getLogger(__name__)


class StatsDetails:
    def __init__(self):
        self.artists = 0
        self.albums = 0
        self.tracks = 0


class Stats:
    def __init__(self):
        self.scanned = 0
        self.added = StatsDetails()
        self.deleted = StatsDetails()
        self.errors = []


class ScanQueue(Queue):
    def _init(self, maxsize):
        self.queue = set()
        self.__last_got = None

    def _put(self, item):
        if self.__last_got != item:
            self.queue.add(item)

    def _get(self):
        self.__last_got = self.queue.pop()
        return self.__last_got


class Scanner(Thread):
    def __init__(
        self,
        force=False,
        extensions=None,
        follow_symlinks=False,
        progress=None,
        on_folder_start=None,
        on_folder_end=None,
        on_done=None,
    ):
        super().__init__()

        if extensions is not None and not isinstance(extensions, list):
            raise TypeError("Invalid extensions type")

        self.__force = force
        self.__extensions = extensions
        self.__follow_symlinks = follow_symlinks

        self.__progress = progress
        self.__on_folder_start = on_folder_start
        self.__on_folder_end = on_folder_end
        self.__on_done = on_done

        self.__stopped = Event()
        self.__queue = ScanQueue()
        self.__stats = Stats()

    scanned = property(lambda self: self.__stats.scanned)

    def __report_progress(self, folder_name, scanned):
        if self.__progress is None:
            return

        self.__progress(folder_name, scanned)

    def queue_folder(self, folder_name):
        if not isinstance(folder_name, str):
            raise TypeError("Expecting string, got " + str(type(folder_name)))

        self.__queue.put(folder_name)

    def run(self):
        opened = open_connection(True)

        while not self.__stopped.is_set():
            try:
                folder_name = self.__queue.get(False)
            except QueueEmpty:
                break

            try:
                folder = Folder.get(name=folder_name, root=True)
            except Folder.DoesNotExist:
                continue

            self.__scan_folder(folder)

        self.prune()

        if self.__on_done is not None:
            self.__on_done()

        if opened:
            close_connection()

    def stop(self):
        self.__stopped.set()

    def __scan_folder(self, folder):
        logger.info("Scanning folder %s", folder.name)

        if self.__on_folder_start is not None:
            self.__on_folder_start(folder)

        # Scan new/updated files
        to_scan = [folder.path]
        scanned = 0
        while not self.__stopped.is_set() and to_scan:
            path = to_scan.pop()
            for entry in os.scandir(path):
                if entry.name.startswith("."):
                    continue
                if entry.is_symlink() and not self.__follow_symlinks:
                    continue
                elif entry.is_dir():
                    to_scan.append(entry.path)
                elif entry.is_file() and self.__check_extension(entry.path):
                    self.scan_file(entry)
                    self.__stats.scanned += 1
                    scanned += 1

                    self.__report_progress(folder.name, scanned)

        # Remove deleted/moved folders
        folders = [folder]
        while not self.__stopped.is_set() and folders:
            f = folders.pop()

            if not f.root and not os.path.isdir(f.path):
                self.__stats.deleted.tracks += f.delete_hierarchy()
                continue

            folders += f.children[:]

        # Remove files that have been deleted
        # Could be more efficient if done when walking on the files
        if not self.__stopped.is_set():
            for track in Track.select().where(Track.root_folder == folder):
                if not os.path.exists(track.path) or not self.__check_extension(
                    track.path
                ):
                    self.remove_file(track.path)

        # Update cover art info
        folders = [folder]
        while not self.__stopped.is_set() and folders:
            f = folders.pop()
            self.find_cover(f.path)
            folders += f.children[:]

        if not self.__stopped.is_set():
            folder.last_scan = int(time.time())
            folder.save()

        if self.__on_folder_end is not None:
            self.__on_folder_end(folder)

    def prune(self):
        if self.__stopped.is_set():
            return

        self.__stats.deleted.albums += Album.prune()
        self.__stats.deleted.artists += Artist.prune()
        Folder.prune()

    def __check_extension(self, path):
        if not self.__extensions:
            return True
        return os.path.splitext(path)[1][1:].lower() in self.__extensions

    def scan_file(self, path_or_direntry):
        if isinstance(path_or_direntry, str):
            path = path_or_direntry

            if not os.path.exists(path):
                return

            basename = os.path.basename(path)
            stat = os.stat(path)
        else:
            path = path_or_direntry.path
            basename = path_or_direntry.name
            stat = path_or_direntry.stat()

        try:
            path.encode("utf-8")  # Test for badly encoded paths
        except UnicodeError:
            self.__stats.errors.append(path)
            return

        mtime = int(stat.st_mtime)

        tr = Track.get_or_none(path=path)
        if tr is not None:
            if not self.__force and not mtime > tr.last_modification:
                return

            tag = self.__try_load_tag(path)
            if tag is None:
                self.remove_file(path)
                return
            trdict = {}
        else:
            tag = self.__try_load_tag(path)
            if tag is None:
                return

            trdict = {"path": path}

        artist = (self.__sanitize_str(tag.artist) or "[unknown]")[:255]
        album = (self.__sanitize_str(tag.album) or "[non-album tracks]")[:255]
        albumartist = (self.__sanitize_str(tag.albumartist) or artist)[:255]

        trdict["disc"] = tag.disc or 1
        trdict["number"] = tag.track or 1
        trdict["title"] = (self.__sanitize_str(tag.title) or basename)[:255]
        trdict["year"] = tag.year
        trdict["genre"] = tag.genre
        trdict["duration"] = int(tag.length)
        trdict["has_art"] = bool(tag.images)

        trdict["bitrate"] = tag.bitrate // 1000
        trdict["last_modification"] = mtime

        tralbum = self.__find_album(albumartist, album)
        trartist = self.__find_artist(artist)

        if tr is None:
            trdict["root_folder"] = self.__find_root_folder(path)
            trdict["folder"] = self.__find_folder(path)
            trdict["album"] = tralbum
            trdict["artist"] = trartist
            trdict["created"] = datetime.fromtimestamp(mtime)

            try:
                Track.create(**trdict)
                self.__stats.added.tracks += 1
            except ValueError:
                # Field validation error
                self.__stats.errors.append(path)
        else:
            if tr.album.id != tralbum.id:
                trdict["album"] = tralbum

            if tr.artist.id != trartist.id:
                trdict["artist"] = trartist

            try:
                for attr, value in trdict.items():
                    setattr(tr, attr, value)
                tr.save()
            except ValueError:
                # Field validation error
                self.__stats.errors.append(path)

    def remove_file(self, path):
        if not isinstance(path, str):
            raise TypeError("Expecting string, got " + str(type(path)))

        try:
            Track.get(path=path).delete_instance(recursive=True)
            self.__stats.deleted.tracks += 1
        except Track.DoesNotExist:
            pass

    def move_file(self, src_path, dst_path):
        if not isinstance(src_path, str):
            raise TypeError("Expecting string, got " + str(type(src_path)))
        if not isinstance(dst_path, str):
            raise TypeError("Expecting string, got " + str(type(dst_path)))

        if src_path == dst_path:
            return

        try:
            tr = Track.get(path=src_path)
        except Track.DoesNotExist:
            return

        try:
            tr_dst = Track.get(path=dst_path)
            root = tr_dst.root_folder
            folder = tr_dst.folder
            self.remove_file(dst_path)
            tr.root_folder = root
            tr.folder = folder
        except Track.DoesNotExist:
            root = self.__find_root_folder(dst_path)
            folder = self.__find_folder(dst_path)
            tr.root_folder = root
            tr.folder = folder
        tr.path = dst_path
        tr.save()

    def find_cover(self, dirpath):
        if not isinstance(dirpath, str):  # pragma: nocover
            raise TypeError("Expecting string, got " + str(type(dirpath)))

        if not os.path.exists(dirpath):
            return

        try:
            folder = Folder.get(path=dirpath)
        except Folder.DoesNotExist:
            return

        album_name = None
        track = folder.tracks.select().first()
        if track is not None:
            album_name = track.album.name

        cover = find_cover_in_folder(folder.path, album_name)
        folder.cover_art = cover.name if cover is not None else None
        folder.save()

    def add_cover(self, path):
        if not isinstance(path, str):  # pragma: nocover
            raise TypeError("Expecting string, got " + str(type(path)))

        try:
            folder = Folder.get(path=os.path.dirname(path))
        except Folder.DoesNotExist:
            return

        cover_name = os.path.basename(path)
        if not folder.cover_art:
            folder.cover_art = cover_name
            folder.save()
        elif folder.cover_art != cover_name:
            album_name = None
            track = folder.tracks.select().first()
            if track is not None:
                album_name = track.album.name

            current_cover = CoverFile(folder.cover_art, album_name)
            new_cover = CoverFile(cover_name, album_name)
            if new_cover.score > current_cover.score:
                folder.cover_art = cover_name
                folder.save()

    def __find_album(self, artist, album):
        ar = self.__find_artist(artist)
        al = ar.albums.where(Album.name == album).first()
        if al:
            return al

        self.__stats.added.albums += 1
        return Album.create(name=album, artist=ar)

    def __find_artist(self, artist):
        try:
            return Artist.get(name=artist)
        except Artist.DoesNotExist:
            self.__stats.added.artists += 1
            return Artist.create(name=artist)

    def __find_root_folder(self, path):
        path = os.path.dirname(path)
        for folder in Folder.select().where(Folder.root):
            if path.startswith(folder.path):
                return folder

        raise Exception(
            "Couldn't find the root folder for '{}'.\nDon't scan files that aren't located in a defined music folder".format(
                path
            )
        )

    def __find_folder(self, path):
        children = []
        drive, _ = os.path.splitdrive(path)
        path = os.path.dirname(path)
        while path not in (drive, "/"):
            try:
                folder = Folder.get(path=path)
                break
            except Folder.DoesNotExist:
                pass

            created = datetime.fromtimestamp(os.path.getmtime(path))
            children.append(
                {
                    "root": False,
                    "name": os.path.basename(path),
                    "path": path,
                    "created": created,
                }
            )
            path = os.path.dirname(path)

        assert folder is not None
        while children:
            folder = Folder.create(parent=folder, **children.pop())

        return folder

    def __try_load_tag(self, path):
        try:
            return mediafile.MediaFile(path)
        except mediafile.UnreadableFileError:
            return None

    def __sanitize_str(self, value):
        if value is None:
            return None
        return value.replace("\x00", "").strip()

    def stats(self):
        return self.__stats



# This file is part of Supysonic.
# Supysonic is a Python implementation of the Subsonic server API.
#
# Copyright (C) 2014-2022 Alban 'spl0k' Féron
#
# Distributed under terms of the GNU AGPLv3 license.

import logging
import os.path
import time

from threading import Thread, Condition, Timer
from watchdog.observers import Observer
from watchdog.events import PatternMatchingEventHandler

from . import covers
from .db import Folder, open_connection, close_connection
from .scanner import Scanner

OP_SCAN = 1
OP_REMOVE = 2
OP_MOVE = 4
FLAG_CREATE = 8
FLAG_COVER = 16

logger = logging.getLogger(__name__)


class SupysonicWatcherEventHandler(PatternMatchingEventHandler):
    def __init__(self, extensions):
        patterns = None
        if extensions:
            patterns = ["*." + e.lower() for e in extensions.split()] + [
                "*" + e for e in covers.EXTENSIONS
            ]
        super().__init__(patterns=patterns, ignore_directories=True)

    def dispatch(self, event):
        try:
            super().dispatch(event)
        except Exception as e:  # pragma: nocover
            logger.critical(e)

    def on_created(self, event):
        logger.debug("File created: '%s'", event.src_path)

        op = OP_SCAN | FLAG_CREATE
        if covers.is_valid_cover(event.src_path):
            op |= FLAG_COVER
        self.queue.put(event.src_path, op)

    def on_deleted(self, event):
        logger.debug("File deleted: '%s'", event.src_path)

        op = OP_REMOVE
        _, ext = os.path.splitext(event.src_path)
        if ext in covers.EXTENSIONS:
            op |= FLAG_COVER
        self.queue.put(event.src_path, op)

    def on_modified(self, event):
        logger.debug("File modified: '%s'", event.src_path)
        op = OP_SCAN
        if covers.is_valid_cover(event.src_path):
            op |= FLAG_COVER
        self.queue.put(event.src_path, op)

    def on_moved(self, event):
        logger.debug("File moved: '%s' -> '%s'", event.src_path, event.dest_path)

        op = OP_MOVE
        _, ext = os.path.splitext(event.src_path)
        if ext in covers.EXTENSIONS:
            op |= FLAG_COVER
        self.queue.put(event.dest_path, op, src_path=event.src_path)


class Event:
    def __init__(self, path, operation, **kwargs):
        if operation & (OP_SCAN | OP_REMOVE) == (OP_SCAN | OP_REMOVE):
            raise Exception("Flags SCAN and REMOVE both set")  # pragma: nocover

        self.__path = path
        self.__time = time.time()
        self.__op = operation
        self.__src = kwargs.get("src_path")

    def set(self, operation, **kwargs):
        if operation & (OP_SCAN | OP_REMOVE) == (OP_SCAN | OP_REMOVE):
            raise Exception("Flags SCAN and REMOVE both set")  # pragma: nocover

        self.__time = time.time()
        if operation & OP_SCAN:
            self.__op &= ~OP_REMOVE
        if operation & OP_REMOVE:
            self.__op &= ~OP_SCAN
        if operation & FLAG_CREATE:
            self.__op &= ~OP_MOVE
        self.__op |= operation

        src_path = kwargs.get("src_path")
        if src_path:
            self.__src = src_path

    @property
    def path(self):
        return self.__path

    @property
    def time(self):
        return self.__time

    @property
    def operation(self):
        return self.__op

    @property
    def src_path(self):
        return self.__src


class ScannerProcessingQueue(Thread):
    def __init__(self, delay):
        super().__init__()

        self.__timeout = delay
        self.__cond = Condition()
        self.__timer = None
        self.__queue = {}
        self.__running = True

    def run(self):
        try:
            self.__run()
        except Exception as e:  # pragma: nocover
            logger.critical(e)
            raise e

    def __run(self):
        while self.__running:
            time.sleep(0.1)

            with self.__cond:
                # Flag might have flipped during sleep. Check it again before waiting
                # See issue #263
                if self.__running:
                    self.__cond.wait()

                if not self.__queue:
                    continue

            logger.debug("Instantiating scanner")
            open_connection()
            scanner = Scanner()

            item = self.__next_item()
            while item:
                if item.operation & FLAG_COVER:
                    self.__process_cover_item(scanner, item)
                else:
                    self.__process_regular_item(scanner, item)

                item = self.__next_item()

            scanner.prune()
            close_connection()
            logger.debug("Freeing scanner")

    def __process_regular_item(self, scanner, item):
        if item.operation & OP_MOVE:
            logger.info("Moving: '%s' -> '%s'", item.src_path, item.path)
            scanner.move_file(item.src_path, item.path)

        if item.operation & OP_SCAN:
            logger.info("Scanning: '%s'", item.path)
            scanner.scan_file(item.path)

        if item.operation & OP_REMOVE:
            logger.info("Removing: '%s'", item.path)
            scanner.remove_file(item.path)

    def __process_cover_item(self, scanner, item):
        if item.operation & OP_SCAN:
            if os.path.isdir(item.path):
                logger.info("Looking for covers: '%s'", item.path)
                scanner.find_cover(item.path)
            else:
                logger.info("Potentially adding cover: '%s'", item.path)
                scanner.add_cover(item.path)

        if item.operation & OP_REMOVE:
            logger.info("Removing cover: '%s'", item.path)
            scanner.find_cover(os.path.dirname(item.path))

        if item.operation & OP_MOVE:
            logger.info("Moving cover: '%s' -> '%s'", item.src_path, item.path)
            scanner.find_cover(os.path.dirname(item.src_path))
            scanner.add_cover(item.path)

    def stop(self):
        with self.__cond:
            self.__running = False
            self.__cond.notify()

    def put(self, path, operation, **kwargs):
        if not self.__running:
            raise RuntimeError("Trying to put an item in a stopped queue")

        with self.__cond:
            if path in self.__queue:
                event = self.__queue[path]
                event.set(operation, **kwargs)
            else:
                event = Event(path, operation, **kwargs)
                self.__queue[path] = event

            if operation & OP_MOVE and kwargs["src_path"] in self.__queue:
                previous = self.__queue[kwargs["src_path"]]
                event.set(previous.operation, src_path=previous.src_path)
                del self.__queue[kwargs["src_path"]]

            if self.__timer:
                self.__timer.cancel()
            self.__timer = Timer(self.__timeout, self.__wakeup)
            self.__timer.start()

    def unschedule_paths(self, basepath):
        with self.__cond:
            for path in list(self.__queue):
                if path.startswith(basepath):
                    del self.__queue[path]

    def __wakeup(self):
        with self.__cond:
            self.__cond.notify()
            self.__timer = None

    def __next_item(self):
        with self.__cond:
            if not self.__queue:
                return None

            next = min(self.__queue.items(), key=lambda i: i[1].time)
            if not self.__running or next[1].time + self.__timeout <= time.time():
                del self.__queue[next[0]]
                return next[1]

            return None


class SupysonicWatcher:
    def __init__(self, config):
        self.__delay = config.DAEMON["wait_delay"]
        self.__handler = SupysonicWatcherEventHandler(config.BASE["scanner_extensions"])

        self.__folders = {}
        self.__queue = None
        self.__observer = None

    def add_folder(self, folder):
        if isinstance(folder, Folder):
            path = folder.path
        elif isinstance(folder, str):
            path = folder
        else:
            raise TypeError("Expecting string or Folder, got " + str(type(folder)))

        logger.info("Scheduling watcher for %s", path)
        watch = self.__observer.schedule(self.__handler, path, recursive=True)
        self.__folders[path] = watch

    def remove_folder(self, folder):
        if isinstance(folder, Folder):
            path = folder.path
        elif isinstance(folder, str):
            path = folder
        else:
            raise TypeError("Expecting string or Folder, got " + str(type(folder)))

        logger.info("Unscheduling watcher for %s", path)
        self.__observer.unschedule(self.__folders[path])
        del self.__folders[path]
        self.__queue.unschedule_paths(path)

    def start(self):
        self.__queue = ScannerProcessingQueue(self.__delay)
        self.__observer = Observer()
        self.__handler.queue = self.__queue

        for folder in Folder.select().where(Folder.root):
            self.add_folder(folder)

        logger.info("Starting watcher")
        self.__queue.start()
        self.__observer.start()

    def stop(self):
        logger.info("Stopping watcher")
        if self.__observer is not None:
            self.__observer.stop()
            self.__observer.join()
        if self.__queue is not None:
            self.__queue.stop()
            self.__queue.join()

        self.__observer = None
        self.__queue = None
        self.__handler.queue = None

    @property
    def running(self):
        return (
            self.__queue is not None
            and self.__observer is not None
            and self.__queue.is_alive()
            and self.__observer.is_alive()
        )



# This file is part of Supysonic.
# Supysonic is a Python implementation of the Subsonic server API.
#
# Copyright (C) 2019 Alban 'spl0k' Féron
#
# Distributed under terms of the GNU AGPLv3 license.


class DaemonUnavailableError(Exception):
    pass



# This file is part of Supysonic.
# Supysonic is a Python implementation of the Subsonic server API.
#
# Copyright (C) 2013-2019 Alban 'spl0k' Féron
#               2018-2019 Carey 'pR0Ps' Metcalfe
#
# Distributed under terms of the GNU AGPLv3 license.

from collections import OrderedDict, namedtuple
import contextlib
import errno
import logging
import os
import os.path
import tempfile
import threading
from time import time


logger = logging.getLogger(__name__)


class CacheMiss(KeyError):
    """The requested data is not in the cache"""

    pass


class ProtectedError(Exception):
    """The data cannot be purged from the cache"""

    pass


CacheEntry = namedtuple("CacheEntry", ["size", "expires"])
NULL_ENTRY = CacheEntry(0, 0)


class Cache:
    """Provides a common interface for caching files to disk"""

    # Modeled after werkzeug.contrib.cache.FileSystemCache

    # keys must be filename-compatible strings (no paths)
    # values must be bytes (not strings)

    def __init__(self, cache_dir, max_size, min_time=300, auto_prune=True):
        """Initialize the cache

        cache_dir: The folder to store cached files
        max_size: The maximum allowed size of the cache in bytes
        min_time: The minimum amount of time a file will be stored in the cache
                  in seconds (default 300 = 5min)
        auto_prune: If True (default) the cache will automatically be pruned to
                    the max_size when possible.

        Note that max_size is not a hard restriction and in some cases will
        temporarily be exceeded, even when auto-pruning is turned on.
        """
        self._cache_dir = os.path.abspath(cache_dir)
        self.min_time = min_time
        self.max_size = max_size
        self._auto_prune = auto_prune
        self._lock = threading.RLock()

        # Create the cache directory
        try:
            os.makedirs(self._cache_dir)
        except OSError as e:
            if e.errno != errno.EEXIST:
                raise

        # Make a key -> CacheEntry(size, expiry) map ordered by mtime
        self._size = 0
        self._files = OrderedDict()
        for mtime, size, key in sorted(
            [
                (f.stat().st_mtime, f.stat().st_size, f.name)
                for f in os.scandir(self._cache_dir)
                if f.is_file()
            ]
        ):
            self._files[key] = CacheEntry(size, mtime + self.min_time)
            self._size += size

    def _filepath(self, key):
        return os.path.join(self._cache_dir, key)

    def _make_space(self, required_space, key=None):
        """Delete files to free up the required space (or close to it)

        If key is provided and exists in the cache, its size will be
        subtracted from the required size.
        """
        target = self.max_size - required_space
        if key is not None:
            target += self._files.get(key, NULL_ENTRY).size

        with self._lock:
            # Delete the oldest file until self._size <= target
            for k in list(self._files.keys()):
                if self._size <= target:
                    break
                try:
                    self.delete(k)
                except ProtectedError:
                    pass

    def _record_file(self, key, size):
        # If the file is being replaced, add only the difference in size
        self._size += size - self._files.get(key, NULL_ENTRY).size
        self._files[key] = CacheEntry(size, int(time()) + self.min_time)

    def _freshen_file(self, key):
        """Touch the file to change modified time and move it to the end of the cache dict"""
        old = self._files.pop(key)
        self._files[key] = CacheEntry(old.size, int(time()) + self.min_time)
        os.utime(self._filepath(key), None)

    @property
    def size(self):
        """The current amount of data cached"""
        return self._size

    def touch(self, key):
        """Mark a cache entry as fresh"""
        with self._lock:
            if not self.has(key):
                raise CacheMiss(key)
            self._freshen_file(key)

    @contextlib.contextmanager
    def set_fileobj(self, key):
        """Yields a file object that can have bytes written to it in order to
        store them in the cache.

        The contents of the file object will be stored in the cache when the
        context is exited.

        Ex:
        >>> with cache.set_fileobj(key) as fp:
        ...     json.dump(some_data, fp)
        """
        f = tempfile.NamedTemporaryFile(
            dir=self._cache_dir, suffix=".part", delete=False
        )
        try:
            yield f

            # seek to end and get position to get filesize
            f.seek(0, 2)
            size = f.tell()
            f.close()

            with self._lock:
                if self._auto_prune:
                    self._make_space(size, key=key)
                os.replace(f.name, self._filepath(key))
                self._record_file(key, size)
        except BaseException:
            f.close()
            with contextlib.suppress(OSError):
                os.remove(f.name)
            raise

    def set(self, key, value):
        """Set a literal value into the cache and return its path"""
        with self.set_fileobj(key) as f:
            f.write(value)
        return self._filepath(key)

    def set_generated(self, key, gen_function):
        """Pass the values yielded from the generator function through and set
        the end result in the cache.

        The contents will be set into the cache only if and when the generator
        completes.

        Ex:
        >>> for x in cache.set_generated(key, generator_function):
        ...     print(x)
        """
        with self.set_fileobj(key) as f:
            gen = gen_function()
            try:
                for data in gen:
                    f.write(data)
                    yield data
            except GeneratorExit:
                # Try to stop the generator but check it still wants to yield data.
                # If it does allow caching of this data without forwarding it
                try:
                    f.write(gen.throw(GeneratorExit))
                    for data in gen:
                        f.write(data)
                except StopIteration:
                    # We stopped just at the end of the generator
                    pass

    def get(self, key):
        """Return the path to the file where the cached data is stored"""
        self.touch(key)
        return self._filepath(key)

    @contextlib.contextmanager
    def get_fileobj(self, key):
        """Yields a file object that can be used to read cached bytes"""
        with open(self.get(key), "rb") as f:
            yield f

    def get_value(self, key):
        """Return the cached data"""
        with self.get_fileobj(key) as f:
            return f.read()

    def delete(self, key):
        """Delete a file from the cache"""
        with self._lock:
            if not self.has(key):
                return
            if time() < self._files[key].expires:
                raise ProtectedError("File has not expired")

            os.remove(self._filepath(key))
            self._size -= self._files.pop(key).size

    def prune(self):
        """Prune the cache down to the max size

        Note that protected files are not deleted
        """
        self._make_space(0)

    def clear(self):
        """Clear the cache

        Note that protected files are not deleted
        """
        self._make_space(self.max_size)

    def has(self, key):
        """Check if a key is currently cached"""
        if key not in self._files:
            return False

        if not os.path.exists(self._filepath(key)):
            # Underlying file is gone, remove from the cache
            self._size -= self._files.pop(key).size
            return False

        return True



# This file is part of Supysonic.
# Supysonic is a Python implementation of the Subsonic server API.
#
# Copyright (C) 2017-2022 Alban 'spl0k' Féron
#
# Distributed under terms of the GNU AGPLv3 license.

import os.path
import unittest
import uuid

from contextlib import closing
from io import BytesIO
from PIL import Image

from supysonic.db import Folder, Artist, Album, Track

from .apitestbase import ApiTestBase


class MediaTestCase(ApiTestBase):
    def setUp(self):
        super().setUp()

        folder = Folder.create(
            name="Root",
            path=os.path.abspath("tests/assets"),
            root=True,
            cover_art="cover.jpg",
        )
        folder = Folder.get(name="Root")
        self.folderid = folder.id

        artist = Artist.create(name="Artist")
        album = Album.create(artist=artist, name="Album")

        track = Track.create(
            title="23bytes",
            number=1,
            disc=1,
            artist=artist,
            album=album,
            path=os.path.abspath("tests/assets/23bytes"),
            root_folder=folder,
            folder=folder,
            duration=2,
            bitrate=320,
            last_modification=0,
        )
        self.trackid = track.id

        self.formats = ["mp3", "flac", "ogg", "m4a"]
        for i in range(len(self.formats)):
            track_embeded_art = Track.create(
                title="[silence]",
                number=1,
                disc=1,
                artist=artist,
                album=album,
                path=os.path.abspath(f"tests/assets/formats/silence.{self.formats[i]}"),
                root_folder=folder,
                folder=folder,
                duration=2,
                bitrate=320,
                last_modification=0,
            )
            self.formats[i] = track_embeded_art.id

    def test_stream(self):
        self._make_request("stream", error=10)
        self._make_request("stream", {"id": "string"}, error=0)
        self._make_request("stream", {"id": str(uuid.uuid4())}, error=70)
        self._make_request("stream", {"id": str(self.folderid)}, error=0)
        self._make_request(
            "stream", {"id": str(self.trackid), "maxBitRate": "string"}, error=0
        )
        self._make_request(
            "stream", {"id": str(self.trackid), "timeOffset": 2}, error=0
        )
        self._make_request(
            "stream", {"id": str(self.trackid), "size": "640x480"}, error=0
        )

        with closing(
            self.client.get(
                "/rest/stream.view",
                query_string={
                    "u": "alice",
                    "p": "Alic3",
                    "c": "tests",
                    "id": str(self.trackid),
                },
            )
        ) as rv:
            self.assertEqual(rv.status_code, 200)
            self.assertEqual(len(rv.data), 23)
        self.assertEqual(Track[self.trackid].play_count, 1)

    def test_download(self):
        self._make_request("download", error=10)
        self._make_request("download", {"id": "string"}, error=0)
        self._make_request("download", {"id": str(uuid.uuid4())}, error=70)

        # download single file
        with closing(
            self.client.get(
                "/rest/download.view",
                query_string={
                    "u": "alice",
                    "p": "Alic3",
                    "c": "tests",
                    "id": str(self.trackid),
                },
            )
        ) as rv:
            self.assertEqual(rv.status_code, 200)
            self.assertEqual(len(rv.data), 23)
        self.assertEqual(Track[self.trackid].play_count, 0)

        # dowload folder
        rv = self.client.get(
            "/rest/download.view",
            query_string={
                "u": "alice",
                "p": "Alic3",
                "c": "tests",
                "id": str(self.folderid),
            },
        )
        self.assertEqual(rv.status_code, 200)
        self.assertEqual(rv.mimetype, "application/zip")

    def __assert_image_data(self, resp, format, size):
        with Image.open(BytesIO(resp.data)) as im:
            self.assertEqual(im.format, format)
            self.assertEqual(im.size, (size, size))

    def test_get_cover_art(self):
        self._make_request("getCoverArt", error=10)
        self._make_request("getCoverArt", {"id": "string"}, error=0)
        self._make_request("getCoverArt", {"id": str(uuid.uuid4())}, error=70)
        self._make_request("getCoverArt", {"id": str(self.trackid)}, error=70)
        self._make_request(
            "getCoverArt", {"id": str(self.folderid), "size": "large"}, error=0
        )

        args = {"u": "alice", "p": "Alic3", "c": "tests", "id": str(self.folderid)}
        with closing(
            self.client.get("/rest/getCoverArt.view", query_string=args)
        ) as rv:
            self.assertEqual(rv.status_code, 200)
            self.assertEqual(rv.mimetype, "image/jpeg")
            self.__assert_image_data(rv, "JPEG", 420)

        args["size"] = 600
        with closing(
            self.client.get("/rest/getCoverArt.view", query_string=args)
        ) as rv:
            self.assertEqual(rv.status_code, 200)
            self.assertEqual(rv.mimetype, "image/jpeg")
            self.__assert_image_data(rv, "JPEG", 420)

        args["size"] = 120
        with closing(
            self.client.get("/rest/getCoverArt.view", query_string=args)
        ) as rv:
            self.assertEqual(rv.status_code, 200)
            self.assertEqual(rv.mimetype, "image/jpeg")
            self.__assert_image_data(rv, "JPEG", 120)

        # rerequest, just in case
        with closing(
            self.client.get("/rest/getCoverArt.view", query_string=args)
        ) as rv:
            self.assertEqual(rv.status_code, 200)
            self.assertEqual(rv.mimetype, "image/jpeg")
            self.__assert_image_data(rv, "JPEG", 120)

        # TODO test non square covers

        # Test extracting cover art from embeded media
        for args["id"] in self.formats:
            with closing(
                self.client.get("/rest/getCoverArt.view", query_string=args)
            ) as rv:
                self.assertEqual(rv.status_code, 200)
                self.assertEqual(rv.mimetype, "image/png")
                self.__assert_image_data(rv, "PNG", 120)

    def test_get_avatar(self):
        self._make_request("getAvatar", error=0)


if __name__ == "__main__":
    unittest.main()



# This file is part of Supysonic.
# Supysonic is a Python implementation of the Subsonic server API.
#
# Copyright (C) 2018 Alban 'spl0k' Féron
#
# Distributed under terms of the GNU AGPLv3 license.

import binascii


def hexlify(s):
    return binascii.hexlify(s.encode("utf-8")).decode("utf-8")



# This file is part of Supysonic.
# Supysonic is a Python implementation of the Subsonic server API.
#
# Copyright (C) 2017-2023 Alban 'spl0k' Féron
#
# Distributed under terms of the GNU AGPLv3 license.

import re
import unittest
import uuid

from collections import namedtuple
from peewee import IntegrityError

from supysonic import db

date_regex = re.compile(r"^\d{4}-\d{2}-\d{2}T\d{2}:\d{2}:\d{2}$")


class DbTestCase(unittest.TestCase):
    def setUp(self):
        db.init_database("sqlite:")

    def tearDown(self):
        db.release_database()

    def create_some_folders(self):
        root_folder = db.Folder.create(root=True, name="Root folder", path="tests")

        f1 = db.Folder.create(
            root=False,
            name="Child folder",
            path="tests/assets",
            cover_art="cover.jpg",
            parent=root_folder,
        )

        f2 = db.Folder.create(
            root=False,
            name="Child folder (No Art)",
            path="tests/formats",
            parent=root_folder,
        )

        return root_folder, f1, f2

    def create_some_tracks(self, artist=None, album=None):
        root, child, child_2 = self.create_some_folders()

        if not artist:
            artist = db.Artist.create(name="Test artist")

        if not album:
            album = db.Album.create(artist=artist, name="Test Album")

        track1 = db.Track.create(
            title="Track Title",
            album=album,
            artist=artist,
            disc=1,
            number=1,
            duration=3,
            has_art=True,
            bitrate=320,
            path="tests/assets/formats/silence.ogg",
            last_modification=1234,
            root_folder=root,
            folder=child,
        )

        track2 = db.Track.create(
            title="One Awesome Song",
            album=album,
            artist=artist,
            disc=1,
            number=2,
            duration=5,
            bitrate=96,
            path="tests/assets/23bytes",
            last_modification=1234,
            root_folder=root,
            folder=child,
        )

        return track1, track2

    def create_track_in(self, folder, root, artist=None, album=None, has_art=True):
        artist = artist or db.Artist.create(name="Snazzy Artist")
        album = album or db.Album.create(artist=artist, name="Rockin' Album")
        return db.Track.create(
            title="Nifty Number",
            album=album,
            artist=artist,
            disc=1,
            number=1,
            duration=5,
            has_art=has_art,
            bitrate=96,
            path="tests/assets/formats/silence.flac",
            last_modification=1234,
            root_folder=root,
            folder=folder,
        )

    def create_user(self, name="Test User"):
        return db.User.create(name=name, password="secret", salt="ABC+")

    def create_playlist(self):
        playlist = db.Playlist.create(user=self.create_user(), name="Playlist!")

        return playlist

    def test_ensure_sqlite_foreign_keys(self):
        root, _, _ = self.create_some_folders()
        self.assertRaises(IntegrityError, root.delete_instance)

    def test_folder_base(self):
        root_folder, child_folder, child_noart = self.create_some_folders()
        track_embededart = self.create_track_in(child_noart, root_folder)

        MockUser = namedtuple("User", ["id"])
        user = MockUser(uuid.uuid4())

        root = root_folder.as_subsonic_child(user)
        self.assertIsInstance(root, dict)
        self.assertIn("id", root)
        self.assertIn("isDir", root)
        self.assertIn("title", root)
        self.assertIn("album", root)
        self.assertIn("created", root)
        self.assertTrue(root["isDir"])
        self.assertEqual(root["title"], "Root folder")
        self.assertEqual(root["album"], "Root folder")
        self.assertRegex(root["created"], date_regex)

        child = child_folder.as_subsonic_child(user)
        self.assertIn("parent", child)
        self.assertIn("artist", child)
        self.assertIn("coverArt", child)
        self.assertEqual(child["parent"], str(root_folder.id))
        self.assertEqual(child["artist"], root_folder.name)
        self.assertEqual(child["coverArt"], child["id"])

        noart = child_noart.as_subsonic_child(user)
        self.assertIn("coverArt", noart)
        self.assertEqual(noart["coverArt"], str(track_embededart.id))

    def test_folder_annotation(self):
        root_folder, child_folder, _ = self.create_some_folders()

        user = self.create_user()
        db.StarredFolder.create(user=user, starred=root_folder)
        db.RatingFolder.create(user=user, rated=root_folder, rating=2)
        other = self.create_user("Other")
        db.RatingFolder.create(user=other, rated=root_folder, rating=5)

        root = root_folder.as_subsonic_child(user)
        self.assertIn("starred", root)
        self.assertIn("userRating", root)
        self.assertIn("averageRating", root)
        self.assertRegex(root["starred"], date_regex)
        self.assertEqual(root["userRating"], 2)
        self.assertEqual(root["averageRating"], 3.5)

        child = child_folder.as_subsonic_child(user)
        self.assertNotIn("starred", child)
        self.assertNotIn("userRating", child)

    def test_artist(self):
        artist = db.Artist.create(name="Test Artist")

        user = self.create_user()
        db.StarredArtist.create(user=user, starred=artist)

        artist_dict = artist.as_subsonic_artist(user)
        self.assertIsInstance(artist_dict, dict)
        self.assertIn("id", artist_dict)
        self.assertIn("name", artist_dict)
        self.assertIn("albumCount", artist_dict)
        self.assertIn("starred", artist_dict)
        self.assertEqual(artist_dict["name"], "Test Artist")
        self.assertEqual(artist_dict["albumCount"], 0)
        self.assertRegex(artist_dict["starred"], date_regex)

        db.Album.create(name="Test Artist", artist=artist)  # self-titled
        db.Album.create(name="The Album After The First One", artist=artist)

        artist_dict = artist.as_subsonic_artist(user)
        self.assertEqual(artist_dict["albumCount"], 2)

    def test_album(self):
        artist = db.Artist.create(name="Test Artist")
        album = db.Album.create(artist=artist, name="Test Album")

        user = self.create_user()
        db.StarredAlbum.create(user=user, starred=album)

        root_folder, folder_art, folder_noart = self.create_some_folders()
        track1 = self.create_track_in(
            folder_noart, root_folder, artist=artist, album=album
        )

        album_dict = album.as_subsonic_album(user)
        self.assertIsInstance(album_dict, dict)
        self.assertIn("id", album_dict)
        self.assertIn("name", album_dict)
        self.assertIn("artist", album_dict)
        self.assertIn("artistId", album_dict)
        self.assertIn("songCount", album_dict)
        self.assertIn("duration", album_dict)
        self.assertIn("created", album_dict)
        self.assertIn("starred", album_dict)
        self.assertIn("coverArt", album_dict)
        self.assertEqual(album_dict["name"], album.name)
        self.assertEqual(album_dict["artist"], artist.name)
        self.assertEqual(album_dict["artistId"], str(artist.id))
        self.assertEqual(album_dict["songCount"], 1)
        self.assertEqual(album_dict["duration"], 5)
        self.assertEqual(album_dict["coverArt"], str(track1.id))
        self.assertRegex(album_dict["created"], date_regex)
        self.assertRegex(album_dict["starred"], date_regex)

    def test_track(self):
        track1, track2 = self.create_some_tracks()

        # Assuming SQLite doesn't enforce foreign key constraints
        MockUser = namedtuple("User", ["id"])
        user = MockUser(uuid.uuid4())

        track1_dict = track1.as_subsonic_child(user, None)
        self.assertIsInstance(track1_dict, dict)
        self.assertIn("id", track1_dict)
        self.assertIn("parent", track1_dict)
        self.assertIn("isDir", track1_dict)
        self.assertIn("title", track1_dict)
        self.assertFalse(track1_dict["isDir"])
        self.assertIn("coverArt", track1_dict)
        self.assertEqual(track1_dict["coverArt"], track1_dict["id"])

        track2_dict = track2.as_subsonic_child(user, None)
        self.assertEqual(track2_dict["coverArt"], track2_dict["parent"])
        # ... we'll test the rest against the API XSD.

    def test_user(self):
        user = self.create_user()

        user_dict = user.as_subsonic_user()
        self.assertIsInstance(user_dict, dict)

    def test_chat(self):
        user = self.create_user()

        line = db.ChatMessage(user=user, message="Hello world!")

        line_dict = line.responsize()
        self.assertIsInstance(line_dict, dict)
        self.assertIn("username", line_dict)
        self.assertEqual(line_dict["username"], user.name)

    def test_playlist(self):
        playlist = self.create_playlist()
        playlist_dict = playlist.as_subsonic_playlist(playlist.user)
        self.assertIsInstance(playlist_dict, dict)

    def test_playlist_tracks(self):
        playlist = self.create_playlist()
        track1, track2 = self.create_some_tracks()

        playlist.add(track1)
        playlist.add(track2)
        self.assertSequenceEqual(playlist.get_tracks(), [track1, track2])

        playlist.add(track2.id)
        playlist.add(track1.id)
        self.assertSequenceEqual(
            playlist.get_tracks(), [track1, track2, track2, track1]
        )

        playlist.clear()
        self.assertSequenceEqual(playlist.get_tracks(), [])

        playlist.add(str(track1.id))
        self.assertSequenceEqual(playlist.get_tracks(), [track1])

        self.assertRaises(ValueError, playlist.add, "some string")
        self.assertRaises(NameError, playlist.add, 2345)

    def test_playlist_remove_tracks(self):
        playlist = self.create_playlist()
        track1, track2 = self.create_some_tracks()

        playlist.add(track1)
        playlist.add(track2)
        playlist.remove_at_indexes([0, 2])
        self.assertSequenceEqual(playlist.get_tracks(), [track2])

        playlist.add(track1)
        playlist.add(track2)
        playlist.add(track2)
        playlist.remove_at_indexes([2, 1])
        self.assertSequenceEqual(playlist.get_tracks(), [track2, track2])

        playlist.add(track1)
        playlist.remove_at_indexes([1, 1])
        self.assertSequenceEqual(playlist.get_tracks(), [track2, track1])

    def test_playlist_fixing(self):
        playlist = self.create_playlist()
        track1, track2 = self.create_some_tracks()

        playlist.add(track1)
        playlist.add(uuid.uuid4())
        playlist.add(track2)
        self.assertSequenceEqual(playlist.get_tracks(), [track1, track2])

        track2.delete_instance()
        self.assertSequenceEqual(playlist.get_tracks(), [track1])

        playlist.tracks = "{0},{0},some random garbage,{0}".format(track1.id)
        self.assertSequenceEqual(playlist.get_tracks(), [track1, track1, track1])


if __name__ == "__main__":
    unittest.main()



# This file is part of Supysonic.
# Supysonic is a Python implementation of the Subsonic server API.
#
# Copyright (C) 2017-2018 Alban 'spl0k' Féron
#
# Distributed under terms of the GNU AGPLv3 license.

import unittest
import uuid

from markupsafe import escape

from supysonic.db import User, ClientPrefs

from .frontendtestbase import FrontendTestBase


class UserTestCase(FrontendTestBase):
    def setUp(self):
        super().setUp()

        self.users = {u.name: u.id for u in User.select()}

    def test_index(self):
        self._login("bob", "B0b")
        rv = self.client.get("/user", follow_redirects=True)
        self.assertIn("There's nothing much to see", rv.data)
        self.assertNotIn("Users", rv.data)
        self._logout()

        self._login("alice", "Alic3")
        rv = self.client.get("/user")
        self.assertIn("Users", rv.data)

    def test_details(self):
        self._login("alice", "Alic3")
        rv = self.client.get("/user/string", follow_redirects=True)
        self.assertIn("badly formed", rv.data)
        rv = self.client.get("/user/" + str(uuid.uuid4()), follow_redirects=True)
        self.assertIn("No such user", rv.data)
        rv = self.client.get("/user/" + str(self.users["bob"]))
        self.assertIn("bob", rv.data)
        self._logout()

        ClientPrefs.create(user=User[self.users["bob"]], client_name="tests")

        self._login("bob", "B0b")
        rv = self.client.get("/user/" + str(self.users["alice"]), follow_redirects=True)
        self.assertIn("There's nothing much to see", rv.data)
        self.assertNotIn("<h2>bob</h2>", rv.data)
        rv = self.client.get("/user/me")
        self.assertIn("<h2>bob</h2>", rv.data)
        self.assertIn("tests", rv.data)

    def test_update_client_prefs(self):
        self._login("alice", "Alic3")
        rv = self.client.post("/user/me")
        self.assertIn("updated", rv.data)  # does nothing, says it's updated anyway
        # error cases, silently ignored
        self.client.post("/user/me", data={"garbage": "trash"})
        self.client.post("/user/me", data={"a_b_c_d_e_f": "g_h_i_j_k"})
        self.client.post("/user/me", data={"_l": "m"})
        self.client.post("/user/me", data={"n_": "o"})
        self.client.post("/user/me", data={"inexisting_client": "setting"})

        ClientPrefs.create(user=User[self.users["alice"]], client_name="tests")

        rv = self.client.post(
            "/user/me", data={"tests_format": "mp3", "tests_bitrate": 128}
        )
        self.assertIn("updated", rv.data)
        prefs = ClientPrefs[User[self.users["alice"]], "tests"]
        self.assertEqual(prefs.format, "mp3")
        self.assertEqual(prefs.bitrate, 128)

        self.client.post("/user/me", data={"tests_delete": 1})
        self.assertEqual(ClientPrefs.select().count(), 0)

    def test_change_username_get(self):
        self._login("bob", "B0b")
        rv = self.client.get("/user/whatever/changeusername", follow_redirects=True)
        self.assertIn("There's nothing much to see", rv.data)
        self._logout()

        self._login("alice", "Alic3")
        rv = self.client.get("/user/whatever/changeusername", follow_redirects=True)
        self.assertIn("badly formed", rv.data)
        rv = self.client.get(
            f"/user/{uuid.uuid4()}/changeusername", follow_redirects=True
        )
        self.assertIn("No such user", rv.data)
        self.client.get("/user/{}/changeusername".format(self.users["bob"]))

    def test_change_username_post(self):
        self._login("alice", "Alic3")
        rv = self.client.post("/user/whatever/changeusername", follow_redirects=True)
        self.assertIn("badly formed", rv.data)
        rv = self.client.post(
            f"/user/{uuid.uuid4()}/changeusername", follow_redirects=True
        )
        self.assertIn("No such user", rv.data)

        path = "/user/{}/changeusername".format(self.users["bob"])
        rv = self.client.post(path, follow_redirects=True)
        self.assertIn("required", rv.data)
        rv = self.client.post(path, data={"user": "bob"}, follow_redirects=True)
        self.assertIn("No changes", rv.data)
        rv = self.client.post(
            path, data={"user": "b0b", "admin": 1}, follow_redirects=True
        )
        self.assertIn("updated", rv.data)
        self.assertIn("b0b", rv.data)
        bob = User[self.users["bob"]]
        self.assertEqual(bob.name, "b0b")
        self.assertTrue(bob.admin)
        rv = self.client.post(path, data={"user": "alice"}, follow_redirects=True)
        self.assertEqual(User[self.users["bob"]].name, "b0b")

    def test_change_mail_get(self):
        self._login("alice", "Alic3")
        self.client.get("/user/me/changemail")
        # whatever

    def test_change_mail_post(self):
        self._login("alice", "Alic3")
        self.client.post("/user/me/changemail")
        # whatever

    def test_change_password_get(self):
        self._login("alice", "Alic3")
        rv = self.client.get("/user/me/changepass")
        self.assertIn("Current password", rv.data)
        rv = self.client.get("/user/{}/changepass".format(self.users["bob"]))
        self.assertNotIn("Current password", rv.data)

    def test_change_password_post(self):
        self._login("alice", "Alic3")
        path = "/user/me/changepass"
        rv = self.client.post(path)
        self.assertIn("required", rv.data)
        rv = self.client.post(path, data={"current": "alice"})
        self.assertIn("required", rv.data)
        rv = self.client.post(path, data={"new": "alice"})
        self.assertIn("required", rv.data)
        rv = self.client.post(path, data={"current": "alice", "new": "alice"})
        self.assertIn("password and its confirmation don", rv.data)
        rv = self.client.post(
            path, data={"current": "alice", "new": "alice", "confirm": "alice"}
        )
        self.assertIn("Wrong password", rv.data)
        self._logout()
        rv = self._login("alice", "Alic3")
        self.assertIn("Logged in", rv.data)
        rv = self.client.post(
            path,
            data={"current": "Alic3", "new": "alice", "confirm": "alice"},
            follow_redirects=True,
        )
        self.assertIn("changed", rv.data)
        self._logout()
        rv = self._login("alice", "alice")
        self.assertIn("Logged in", rv.data)

        path = "/user/{}/changepass".format(self.users["bob"])
        rv = self.client.post(path)
        self.assertIn("required", rv.data)
        rv = self.client.post(path, data={"new": "alice"})
        self.assertIn("password and its confirmation don", rv.data)
        rv = self.client.post(
            path, data={"new": "alice", "confirm": "alice"}, follow_redirects=True
        )
        self.assertIn("changed", rv.data)
        self._logout()
        rv = self._login("bob", "alice")
        self.assertIn("Logged in", rv.data)

    def test_add_get(self):
        self._login("bob", "B0b")
        rv = self.client.get("/user/add", follow_redirects=True)
        self.assertIn("There's nothing much to see", rv.data)
        self.assertNotIn("Add User", rv.data)
        self._logout()

        self._login("alice", "Alic3")
        rv = self.client.get("/user/add")
        self.assertIn("Add User", rv.data)

    def test_add_post(self):
        self._login("alice", "Alic3")
        rv = self.client.post("/user/add")
        self.assertIn("required", rv.data)
        rv = self.client.post("/user/add", data={"user": "user"})
        self.assertIn("Please provide a password", rv.data)
        rv = self.client.post("/user/add", data={"passwd": "passwd"})
        self.assertIn("required", rv.data)
        rv = self.client.post("/user/add", data={"user": "name", "passwd": "passwd"})
        self.assertIn("passwords don", rv.data)
        rv = self.client.post(
            "/user/add",
            data={"user": "alice", "passwd": "passwd", "passwd_confirm": "passwd"},
        )
        self.assertIn(escape("User 'alice' exists"), rv.data)
        self.assertEqual(User.select().count(), 2)

        rv = self.client.post(
            "/user/add",
            data={
                "user": "user",
                "passwd": "passwd",
                "passwd_confirm": "passwd",
                "admin": 1,
            },
            follow_redirects=True,
        )
        self.assertIn("added", rv.data)
        self.assertEqual(User.select().count(), 3)
        self._logout()
        rv = self._login("user", "passwd")
        self.assertIn("Logged in", rv.data)

    def test_delete(self):
        path = "/user/del/{}".format(self.users["bob"])

        self._login("bob", "B0b")
        rv = self.client.get(path, follow_redirects=True)
        self.assertIn("There's nothing much to see", rv.data)
        self.assertEqual(User.select().count(), 2)
        self._logout()

        self._login("alice", "Alic3")
        rv = self.client.get("/user/del/string", follow_redirects=True)
        self.assertIn("badly formed", rv.data)
        rv = self.client.get("/user/del/" + str(uuid.uuid4()), follow_redirects=True)
        self.assertIn("No such user", rv.data)
        rv = self.client.get(path, follow_redirects=True)
        self.assertIn("Deleted", rv.data)
        self.assertEqual(User.select().count(), 1)
        self._logout()
        rv = self._login("bob", "B0b")
        self.assertIn("Wrong username or password", rv.data)

    def test_lastfm_link(self):
        self._login("alice", "Alic3")
        rv = self.client.get("/user/me/lastfm/link", follow_redirects=True)
        self.assertIn("Missing LastFM auth token", rv.data)
        rv = self.client.get(
            "/user/me/lastfm/link",
            query_string={"token": "abcdef"},
            follow_redirects=True,
        )
        self.assertIn("No API key set", rv.data)

    def test_lastfm_unlink(self):
        self._login("alice", "Alic3")
        rv = self.client.get("/user/me/lastfm/unlink", follow_redirects=True)
        self.assertIn("Unlinked", rv.data)

    def test_listenbrainz_unlink(self):
        self._login("alice", "Alic3")
        rv = self.client.get("/user/me/listenbrainz/unlink", follow_redirects=True)
        self.assertIn("Unlinked", rv.data)


if __name__ == "__main__":
    unittest.main()



# This file is part of Supysonic.
# Supysonic is a Python implementation of the Subsonic server API.
#
# Copyright (C) 2017-2019 Alban 'spl0k' Féron
#               2017 Óscar García Amor
#
# Distributed under terms of the GNU AGPLv3 license.

import os.path


def load_tests(loader, tests, pattern):
    this_dir = os.path.dirname(__file__)
    tests.addTests(loader.discover(start_dir=this_dir, pattern="test*.py"))
    tests.addTests(loader.discover(start_dir=this_dir, pattern="issue*.py"))
    return tests



import os.path

# Simulate import of the "supysonic" package
supy_module_path = os.path.join(
    os.path.dirname(__file__), "..", "supysonic", "__init__.py"
)
with open(supy_module_path, "rt", encoding="utf-8") as f:
    supysonic = type("", (), {})()
    exec(f.read(), supysonic.__dict__)


# -- Project information -----------------------------------------------------

project = supysonic.NAME
author = supysonic.AUTHOR
copyright = "2013-2023, " + author

version = supysonic.VERSION
release = supysonic.VERSION


# -- General configuration ---------------------------------------------------

extensions = []
templates_path = []
source_suffix = ".rst"
master_doc = "index"
exclude_patterns = ["_build", "Thumbs.db", ".DS_Store"]
smartquotes_action = "qe"

primary_domain = None
highlight_language = "none"

language = None


# -- Options for HTML output -------------------------------------------------

html_theme = "alabaster"
html_theme_options = {
    "description": supysonic.DESCRIPTION,
    "github_user": "spl0k",
    "github_repo": "supysonic",
}
html_static_path = ["_static"]

html_sidebars = {
    "*": [
        "about.html",
        "navigation.html",
        "relations.html",
        "searchbox.html",
        "donate.html",
    ],
    "setup/**": [
        "about.html",
        "localtoc.html",
        "navigation.html",
        "relations.html",
        "searchbox.html",
        "donate.html",
    ],
}

html_domain_indices = False


# -- Options for manual page output ------------------------------------------

_man_authors = ["Louis-Philippe Véronneau", author]

# (source start file, name, description, authors, manual section).
man_pages = [
    (
        "man/supysonic-cli",
        "supysonic-cli",
        "Supysonic management command line interface",
        _man_authors,
        1,
    ),
    (
        "man/supysonic-cli-user",
        "supysonic-cli-user",
        "Supysonic user management commands",
        _man_authors,
        1,
    ),
    (
        "man/supysonic-cli-folder",
        "supysonic-cli-folder",
        "Supysonic folder management commands",
        _man_authors,
        1,
    ),
    (
        "man/supysonic-daemon",
        "supysonic-daemon",
        "Supysonic background daemon",
        _man_authors,
        1,
    ),
    (
        "man/supysonic-server",
        "supysonic-server",
        "Python implementation of the Subsonic server API",
        [author],
        1,
    ),
]



#
# This file is part of Supysonic.
# Supysonic is a Python implementation of the Subsonic server API.
#
# Copyright (C) 2013-2018 Alban 'spl0k' Féron
#
# Distributed under terms of the GNU AGPLv3 license.

from flask import request

from . import api_routing


@api_routing("/ping")
def ping():
    return request.formatter.empty


@api_routing("/getLicense")
def license():
    return request.formatter("license", {"valid": True})



# This file is part of Supysonic.
# Supysonic is a Python implementation of the Subsonic server API.
#
# Copyright (C) 2019-2022 Alban 'spl0k' Féron
#
# Distributed under terms of the GNU AGPLv3 license.

import logging
import shlex
import time

from datetime import datetime, timedelta
from random import shuffle
from subprocess import Popen, DEVNULL
from threading import Thread, Event, RLock

from .db import Track

logger = logging.getLogger(__name__)


class Jukebox:
    def __init__(self, cmd):
        self.__cmd = shlex.split(cmd)
        self.__playlist = []
        self.__index = 0
        self.__offset = 0
        self.__start = None

        self.__thread = None
        self.__lock = RLock()
        self.__skip = Event()
        self.__stop = Event()

    playing = property(
        lambda self: self.__thread is not None and self.__thread.is_alive()
    )
    index = property(lambda self: self.__index)
    gain = property(lambda self: 1.0)
    playlist = property(lambda self: list(self.__playlist))

    @property
    def position(self):
        if self.__start is None:
            return 0
        return int((datetime.utcnow() - self.__start).total_seconds())

    def set(self, *tracks):
        self.clear()
        self.add(*tracks)

    def start(self):
        if self.playing or not self.__playlist:
            return

        self.__skip.clear()
        self.__stop.clear()
        self.__offset = 0
        self.__thread = Thread(target=self.__play_thread)
        self.__thread.start()

    def stop(self):
        if not self.playing:
            return

        self.__stop.set()

    def skip(self, index, offset):
        if index < 0 or index >= len(self.__playlist):
            raise IndexError()
        if offset < 0:
            raise ValueError()

        with self.__lock:
            self.__index = index
            self.__offset = offset
            self.__start = datetime.utcnow() - timedelta(seconds=offset)
        self.__skip.set()
        self.start()

    def add(self, *tracks):
        with self.__lock:
            for t in tracks:
                try:
                    self.__playlist.append(Track[t].path)
                except Track.DoesNotExist:
                    pass

    def clear(self):
        with self.__lock:
            self.__playlist.clear()
            self.__index = 0
            self.__offset = 0

    def remove(self, index):
        try:
            with self.__lock:
                self.__playlist.pop(index)
                if index < self.__index:
                    self.__index -= 1
        except IndexError:
            pass

    def shuffle(self):
        with self.__lock:
            shuffle(self.__playlist)

    def setgain(self, gain):
        pass

    def terminate(self):
        self.__stop.set()
        if self.__thread is not None:
            self.__thread.join()

    def __play_thread(self):
        proc = None
        while not self.__stop.is_set():
            if self.__skip.is_set():
                proc.terminate()
                proc.wait()
                proc = None
                self.__skip.clear()

            if proc is None:
                with self.__lock:
                    proc = self.__play_file()
            elif proc.poll() is not None:
                with self.__lock:
                    self.__start = None
                    self.__index += 1
                    if self.__index >= len(self.__playlist):
                        break

                    proc = self.__play_file()

            time.sleep(0.1)

        proc.terminate()
        proc.wait()
        self.__start = None

    def __play_file(self):
        path = self.__playlist[self.__index]
        args = [
            a.replace("%path", path).replace("%offset", str(self.__offset))
            for a in self.__cmd
        ]

        self.__start = datetime.utcnow() - timedelta(seconds=self.__offset)
        self.__offset = 0

        logger.debug("Start playing with command %s", args)
        try:
            return Popen(args, stdin=DEVNULL, stdout=DEVNULL, stderr=DEVNULL)
        except Exception:
            logger.exception("Failed to run play command")
            return None



# This file is part of Supysonic.
# Supysonic is a Python implementation of the Subsonic server API.
#
# Copyright (C) 2017-2022 Alban 'spl0k' Féron
#
# Distributed under terms of the GNU AGPLv3 license.

import mutagen
import os
import shutil
import tempfile
import time
import unittest

from hashlib import sha1

from supysonic.db import init_database, release_database, Track, Artist, Folder
from supysonic.managers.folder import FolderManager
from supysonic.watcher import SupysonicWatcher

from ..testbase import TestConfig


class WatcherTestConfig(TestConfig):
    DAEMON = {"wait_delay": 0.5, "log_file": "/dev/null", "log_level": "DEBUG"}

    def __init__(self, db_uri):
        super().__init__(False, False)
        self.BASE["database_uri"] = db_uri


class WatcherTestBase(unittest.TestCase):
    def setUp(self):
        self.__db = tempfile.mkstemp()
        dburi = "sqlite:///" + self.__db[1]
        init_database(dburi)

        conf = WatcherTestConfig(dburi)
        self.__sleep_time = conf.DAEMON["wait_delay"] + 1

        self.__watcher = SupysonicWatcher(conf)

    def tearDown(self):
        release_database()
        os.close(self.__db[0])
        os.remove(self.__db[1])

    def _start(self):
        self.__watcher.start()
        time.sleep(0.2)

    def _stop(self):
        self.__watcher.stop()

    def _is_alive(self):
        return self.__watcher.running

    def _sleep(self):
        time.sleep(self.__sleep_time)


class WatcherTestCase(WatcherTestBase):
    def setUp(self):
        super().setUp()
        self.__dir = tempfile.mkdtemp()
        FolderManager.add("Folder", self.__dir)
        self._start()

    def tearDown(self):
        self._stop()
        shutil.rmtree(self.__dir)
        super().tearDown()

    @staticmethod
    def _tempname():
        with tempfile.NamedTemporaryFile() as f:
            return os.path.basename(f.name)

    def _temppath(self, suffix, depth=0):
        if depth > 0:
            dirpath = os.path.join(
                self.__dir, *(self._tempname() for _ in range(depth))
            )
            os.makedirs(dirpath)
        else:
            dirpath = self.__dir
        return os.path.join(dirpath, self._tempname() + suffix)

    def _addfile(self, depth=0):
        path = self._temppath(".mp3", depth)
        shutil.copyfile("tests/assets/folder/silence.mp3", path)
        return path

    def _addcover(self, suffix=None, depth=0):
        suffix = ".jpg" if suffix is None else (suffix + ".jpg")
        path = self._temppath(suffix, depth)
        shutil.copyfile("tests/assets/cover.jpg", path)
        return path


class AudioWatcherTestCase(WatcherTestCase):
    def assertTrackCountEqual(self, expected):
        self.assertEqual(Track.select().count(), expected)

    def test_add(self):
        self._addfile()
        self.assertTrackCountEqual(0)
        self._sleep()
        self.assertTrackCountEqual(1)

    def test_add_nowait_stop(self):
        self._addfile()
        # Add a small delay (< wait_delay) so wathdog can pick up that a file was added
        time.sleep(0.1)
        self._stop()
        self.assertTrackCountEqual(1)

    def test_add_multiple(self):
        self._addfile()
        self._addfile()
        self._addfile()
        self.assertTrackCountEqual(0)
        self._sleep()

        self.assertEqual(Track.select().count(), 3)
        self.assertEqual(Artist.select().count(), 1)

    def test_change(self):
        path = self._addfile()
        self._sleep()

        trackid = None
        self.assertEqual(Track.select().count(), 1)
        self.assertEqual(Artist.select().where(Artist.name == "Some artist").count(), 1)
        trackid = Track.select().first().id

        tags = mutagen.File(path, easy=True)
        tags["artist"] = "Renamed"
        tags.save()
        self._sleep()

        self.assertEqual(Track.select().count(), 1)
        self.assertEqual(Artist.select().where(Artist.name == "Some artist").count(), 0)
        self.assertEqual(Artist.select().where(Artist.name == "Renamed").count(), 1)
        self.assertEqual(Track.select().first().id, trackid)

    def test_rename(self):
        path = self._addfile()
        self._sleep()

        trackid = None
        self.assertEqual(Track.select().count(), 1)
        trackid = Track.select().first().id

        newpath = self._temppath(".mp3")
        shutil.move(path, newpath)
        self._sleep()

        track = Track.select().first()
        self.assertIsNotNone(track)
        self.assertNotEqual(track.path, path)
        self.assertEqual(track.path, newpath)
        self.assertEqual(
            track._path_hash, memoryview(sha1(newpath.encode("utf-8")).digest())
        )
        self.assertEqual(track.id, trackid)

    def test_move_in(self):
        filename = self._tempname() + ".mp3"
        initialpath = os.path.join(tempfile.gettempdir(), filename)
        shutil.copyfile("tests/assets/folder/silence.mp3", initialpath)
        shutil.move(initialpath, self._temppath(".mp3"))
        self._sleep()
        self.assertTrackCountEqual(1)

    def test_move_out(self):
        initialpath = self._addfile()
        self._sleep()
        self.assertTrackCountEqual(1)

        newpath = os.path.join(tempfile.gettempdir(), os.path.basename(initialpath))
        shutil.move(initialpath, newpath)
        self._sleep()
        self.assertTrackCountEqual(0)

        os.unlink(newpath)

    def test_delete(self):
        path = self._addfile()
        self._sleep()
        self.assertTrackCountEqual(1)

        os.unlink(path)
        self._sleep()
        self.assertTrackCountEqual(0)

    def test_add_delete(self):
        path = self._addfile()
        os.unlink(path)
        self._sleep()
        self.assertTrackCountEqual(0)

    def test_add_rename(self):
        path = self._addfile()
        shutil.move(path, self._temppath(".mp3"))
        self._sleep()
        self.assertTrackCountEqual(1)

    def test_rename_delete(self):
        path = self._addfile()
        self._sleep()
        self.assertTrackCountEqual(1)

        newpath = self._temppath(".mp3")
        shutil.move(path, newpath)
        os.unlink(newpath)
        self._sleep()
        self.assertTrackCountEqual(0)

    def test_add_rename_delete(self):
        path = self._addfile()
        newpath = self._temppath(".mp3")
        shutil.move(path, newpath)
        os.unlink(newpath)
        self._sleep()
        self.assertTrackCountEqual(0)

    def test_rename_rename(self):
        path = self._addfile()
        self._sleep()
        self.assertTrackCountEqual(1)

        newpath = self._temppath(".mp3")
        finalpath = self._temppath(".mp3")
        shutil.move(path, newpath)
        shutil.move(newpath, finalpath)
        self._sleep()
        self.assertTrackCountEqual(1)


class CoverWatcherTestCase(WatcherTestCase):
    def test_add_file_then_cover(self):
        self._addfile()
        path = self._addcover()
        self._sleep()

        self.assertEqual(Folder.select().first().cover_art, os.path.basename(path))

    def test_add_cover_then_file(self):
        path = self._addcover()
        self._addfile()
        self._sleep()

        self.assertEqual(Folder.select().first().cover_art, os.path.basename(path))

    def test_remove_cover(self):
        self._addfile()
        path = self._addcover()
        self._sleep()

        os.unlink(path)
        self._sleep()

        self.assertIsNone(Folder.select().first().cover_art)

    def test_naming_add_good(self):
        self._addcover()
        self._sleep()
        good = os.path.basename(self._addcover("cover"))
        self._sleep()

        self.assertEqual(Folder.select().first().cover_art, good)

    def test_naming_add_bad(self):
        good = os.path.basename(self._addcover("cover"))
        self._sleep()
        self._addcover()
        self._sleep()

        self.assertEqual(Folder.select().first().cover_art, good)

    def test_naming_remove_good(self):
        bad = self._addcover()
        good = self._addcover("cover")
        self._sleep()
        os.unlink(good)
        self._sleep()

        self.assertEqual(Folder.select().first().cover_art, os.path.basename(bad))

    def test_naming_remove_bad(self):
        bad = self._addcover()
        good = self._addcover("cover")
        self._sleep()
        os.unlink(bad)
        self._sleep()

        self.assertEqual(Folder.select().first().cover_art, os.path.basename(good))

    def test_rename(self):
        path = self._addcover()
        self._sleep()
        newpath = self._temppath(".jpg")
        shutil.move(path, newpath)
        self._sleep()

        self.assertEqual(Folder.select().first().cover_art, os.path.basename(newpath))

    def test_add_to_folder_without_track(self):
        path = self._addcover(depth=1)
        self._sleep()

        self.assertFalse(
            Folder.select().where(Folder.cover_art == os.path.basename(path)).exists()
        )

    def test_remove_from_folder_without_track(self):
        path = self._addcover(depth=1)
        self._sleep()
        os.unlink(path)
        self._sleep()

    def test_add_track_to_empty_folder(self):
        self._addfile(1)
        self._sleep()


if __name__ == "__main__":
    unittest.main()



# This file is part of Supysonic.
# Supysonic is a Python implementation of the Subsonic server API.
#
# Copyright (C) 2018-2020 Alban 'spl0k' Féron
#
# Distributed under terms of the GNU AGPLv3 license.

import os.path
import re
import warnings

from PIL import Image
from os import scandir


EXTENSIONS = (".jpg", ".jpeg", ".png", ".bmp")
NAMING_SCORE_RULES = (
    ("cover", 5),
    ("albumart", 5),
    ("folder", 5),
    ("front", 10),
    ("back", -10),
    ("large", 2),
    ("small", -2),
)


class CoverFile:
    __clean_regex = re.compile(r"[^a-z]")

    @staticmethod
    def __clean_name(name):
        return CoverFile.__clean_regex.sub("", name.lower())

    def __init__(self, name, album_name=None):
        self.name = name
        self.score = 0

        for part, score in NAMING_SCORE_RULES:
            if part in name.lower():
                self.score += score

        if album_name:
            basename, _ = os.path.splitext(name)
            clean = CoverFile.__clean_name(basename)
            album_name = CoverFile.__clean_name(album_name)
            if clean in album_name or album_name in clean:
                self.score += 20


def is_valid_cover(path):
    if not os.path.isfile(path):
        return False

    _, ext = os.path.splitext(path)
    if ext.lower() not in EXTENSIONS:
        return False

    try:  # Ensure the image can be read
        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            with Image.open(path):
                return True
    except OSError:
        return False


def find_cover_in_folder(path, album_name=None):
    if not os.path.isdir(path):
        raise ValueError("Invalid path")

    candidates = []
    for entry in scandir(path):
        if not is_valid_cover(entry.path):
            continue

        cover = CoverFile(entry.name, album_name)
        candidates.append(cover)

    if not candidates:
        return None
    if len(candidates) == 1:
        return candidates[0]

    return sorted(candidates, key=lambda c: c.score, reverse=True)[0]



# This file is part of Supysonic.
# Supysonic is a Python implementation of the Subsonic server API.
#
# Copyright (C) 2013-2018 Alban 'spl0k' Féron
#                    2017 Óscar García Amor
#
# Distributed under terms of the GNU AGPLv3 license.

NAME = "Supysonic"
VERSION = "0.7.8"
DESCRIPTION = "Python implementation of the Subsonic server API"
AUTHOR = "Alban Féron"
AUTHOR_EMAIL = "alban.feron@gmail.com"
URL = "https://supysonic.readthedocs.io/"
DOWNLOAD_URL = "https://github.com/spl0k/supysonic"
LICENSE = "GNU AGPLv3"



# This file is part of Supysonic.
# Supysonic is a Python implementation of the Subsonic server API.
#
# Copyright (C) 2017-2023 Alban 'spl0k' Féron
#
# Distributed under terms of the GNU AGPLv3 license.

import unittest

from supysonic.db import (
    Folder,
    Artist,
    Album,
    Track,
    StarredArtist,
    StarredAlbum,
    StarredFolder,
    StarredTrack,
    User,
)

from .apitestbase import ApiTestBase


class AlbumSongsTestCase(ApiTestBase):
    # I'm too lazy to write proper tests concerning the data on those endpoints
    # Let's just check paramter validation and ensure coverage

    def setUp(self):
        super().setUp()

        folder = Folder.create(name="Root", root=True, path="tests/assets")
        empty = Folder.create(name="Root", root=True, path="/tmp")
        artist = Artist.create(name="Artist")
        album = Album.create(name="Album", artist=artist)

        Track.create(
            title="Track 1",
            album=album,
            artist=artist,
            disc=1,
            number=1,
            year=123,
            path="tests/assets/folder/1",
            folder=folder,
            root_folder=folder,
            duration=2,
            bitrate=320,
            last_modification=0,
        )
        Track.create(
            title="Track 2",
            album=album,
            artist=artist,
            disc=1,
            number=1,
            year=124,
            genre="Lampshade",
            path="tests/assets/folder/2",
            folder=folder,
            root_folder=folder,
            duration=2,
            bitrate=320,
            last_modification=0,
        )

    def test_get_album_list(self):
        self._make_request("getAlbumList", error=10)
        self._make_request("getAlbumList", {"type": "kraken"}, error=0)
        self._make_request("getAlbumList", {"type": "random", "size": "huge"}, error=0)
        self._make_request(
            "getAlbumList", {"type": "newest", "offset": "minus one"}, error=0
        )
        self._make_request("getAlbumList", {"type": "byYear"}, error=10)
        self._make_request(
            "getAlbumList",
            {"type": "byYear", "fromYear": "Epoch", "toYear": "EOL"},
            error=0,
        )
        self._make_request("getAlbumList", {"type": "byGenre"}, error=10)
        self._make_request(
            "getAlbumList", {"type": "random", "musicFolderId": "id"}, error=0
        )
        self._make_request(
            "getAlbumList", {"type": "random", "musicFolderId": 12}, error=70
        )

        types_and_count = [
            ("random", 1),
            ("newest", 1),
            ("highest", 1),
            ("frequent", 1),
            ("recent", 0),  # never played
            ("alphabeticalByName", 1),
            (
                "alphabeticalByArtist",
                0,  # somehow expected due to funky "album" definition on this endpoint
            ),
            ("starred", 0),  # nothing's starred
        ]
        for t, c in types_and_count:
            rv, child = self._make_request(
                "getAlbumList", {"type": t}, tag="albumList", skip_post=t == "random"
            )
            self.assertEqual(len(child), c)

        rv, child = self._make_request(
            "getAlbumList",
            {"type": "byYear", "fromYear": 100, "toYear": 200},
            tag="albumList",
        )
        self.assertEqual(len(child), 1)
        rv, child = self._make_request(
            "getAlbumList",
            {"type": "byYear", "fromYear": 200, "toYear": 300},
            tag="albumList",
        )
        self.assertEqual(len(child), 0)
        # Need more data to properly test ordering
        rv, child = self._make_request(
            "getAlbumList",
            {"type": "byYear", "fromYear": 200, "toYear": 100},
            tag="albumList",
        )
        self.assertEqual(len(child), 1)

        rv, child = self._make_request(
            "getAlbumList", {"type": "byGenre", "genre": "FARTS"}, tag="albumList"
        )
        self.assertEqual(len(child), 0)

        rv, child = self._make_request(
            "getAlbumList", {"type": "byGenre", "genre": "Lampshade"}, tag="albumList"
        )
        self.assertEqual(len(child), 1)

        _, child = self._make_request(
            "getAlbumList",
            {"musicFolderId": 1, "type": "alphabeticalByName"},
            tag="albumList",
        )
        self.assertEqual(len(child), 1)
        _, child = self._make_request(
            "getAlbumList",
            {"musicFolderId": 2, "type": "alphabeticalByName"},
            tag="albumList",
        )
        self.assertEqual(len(child), 0)

        Track.delete().execute()
        Folder[1].delete_instance()
        rv, child = self._make_request(
            "getAlbumList", {"type": "random"}, tag="albumList"
        )
        self.assertEqual(len(child), 0)

    def test_get_album_list2(self):
        self._make_request("getAlbumList2", error=10)
        self._make_request("getAlbumList2", {"type": "void"}, error=0)
        self._make_request(
            "getAlbumList2", {"type": "random", "size": "size_t"}, error=0
        )
        self._make_request(
            "getAlbumList2", {"type": "newest", "offset": "&v + 2"}, error=0
        )
        self._make_request("getAlbumList2", {"type": "byYear"}, error=10)
        self._make_request(
            "getAlbumList2",
            {"type": "byYear", "fromYear": "Epoch", "toYear": "EOL"},
            error=0,
        )
        self._make_request("getAlbumList2", {"type": "byGenre"}, error=10)
        self._make_request(
            "getAlbumList2", {"type": "random", "musicFolderId": "id"}, error=0
        )
        self._make_request(
            "getAlbumList2", {"type": "random", "musicFolderId": 12}, error=70
        )

        types = [
            "random",
            "newest",
            "frequent",
            "recent",
            "starred",
            "alphabeticalByName",
            "alphabeticalByArtist",
        ]
        for t in types:
            self._make_request(
                "getAlbumList2", {"type": t}, tag="albumList2", skip_post=t == "random"
            )

        self._make_request(
            "getAlbumList2", {"type": "random"}, tag="albumList2", skip_post=True
        )

        rv, child = self._make_request(
            "getAlbumList2",
            {"type": "byYear", "fromYear": 100, "toYear": 200},
            tag="albumList2",
        )
        self.assertEqual(len(child), 1)
        rv, child = self._make_request(
            "getAlbumList2",
            {"type": "byYear", "fromYear": 200, "toYear": 300},
            tag="albumList2",
        )
        self.assertEqual(len(child), 0)
        # Need more data to properly test ordering
        rv, child = self._make_request(
            "getAlbumList2",
            {"type": "byYear", "fromYear": 200, "toYear": 100},
            tag="albumList2",
        )
        self.assertEqual(len(child), 1)

        rv, child = self._make_request(
            "getAlbumList2", {"type": "byGenre", "genre": "FARTS"}, tag="albumList2"
        )
        self.assertEqual(len(child), 0)

        rv, child = self._make_request(
            "getAlbumList2", {"type": "byGenre", "genre": "Lampshade"}, tag="albumList2"
        )
        self.assertEqual(len(child), 1)

        _, child = self._make_request(
            "getAlbumList2",
            {"musicFolderId": 1, "type": "alphabeticalByName"},
            tag="albumList2",
        )
        self.assertEqual(len(child), 1)
        _, child = self._make_request(
            "getAlbumList2",
            {"musicFolderId": 2, "type": "alphabeticalByName"},
            tag="albumList2",
        )
        self.assertEqual(len(child), 0)

        Track.delete().execute()
        Album.delete().execute()
        rv, child = self._make_request(
            "getAlbumList2", {"type": "random"}, tag="albumList2"
        )
        self.assertEqual(len(child), 0)

    def test_get_random_songs(self):
        self._make_request("getRandomSongs", {"size": "8 floors"}, error=0)
        self._make_request("getRandomSongs", {"fromYear": "year"}, error=0)
        self._make_request("getRandomSongs", {"toYear": "year"}, error=0)
        self._make_request("getRandomSongs", {"musicFolderId": "idid"}, error=0)
        self._make_request("getRandomSongs", {"musicFolderId": 1234567890}, error=70)

        rv, child = self._make_request(
            "getRandomSongs", tag="randomSongs", skip_post=True
        )

        self._make_request(
            "getRandomSongs",
            {
                "fromYear": -52,
                "toYear": "1984",
                "genre": "some cryptic subgenre youve never heard of",
                "musicFolderId": 1,
            },
            tag="randomSongs",
        )

    def test_now_playing(self):
        self._make_request("getNowPlaying", tag="nowPlaying")

    def _create_starred_info(self):
        user = User.select().first()
        StarredArtist.create(user=user, starred=Artist.select().first())
        StarredAlbum.create(user=user, starred=Album.select().first())
        StarredTrack.create(user=user, starred=Track.select().first())
        StarredFolder.create(user=user, starred=Folder.select().first())

    def test_get_starred(self):
        self._create_starred_info()

        self._make_request("getStarred", tag="starred")
        self._make_request("getStarred", {"musicFolderId": 1}, tag="starred")

    def test_get_starred2(self):
        self._create_starred_info()

        self._make_request("getStarred2", tag="starred2")
        self._make_request("getStarred2", {"musicFolderId": 1}, tag="starred2")

    def test_get_songs_by_genre(self):
        self._make_request("getSongsByGenre", error=10)
        self._make_request(
            "getSongsByGenre", {"genre": "genre", "musicFolderId": "idid"}, error=0
        )
        self._make_request(
            "getSongsByGenre", {"genre": "genre", "musicFolderId": 1234567890}, error=70
        )
        self._make_request(
            "getSongsByGenre", {"genre": "genre", "count": "three"}, error=0
        )
        self._make_request(
            "getSongsByGenre", {"genre": "genre", "offset": "four"}, error=0
        )

        rv, child = self._make_request(
            "getSongsByGenre", {"genre": "Lampshade"}, tag="songsByGenre"
        )
        self.assertEqual(len(child), 1)


if __name__ == "__main__":
    unittest.main()



# This file is part of Supysonic.
# Supysonic is a Python implementation of the Subsonic server API.
#
# Copyright (C) 2020-2022 Alban 'spl0k' Féron
#
# Distributed under terms of the GNU AGPLv3 license.

from time import sleep
from threading import Thread

from supysonic.daemon.server import Daemon
from supysonic.db import Folder

from .apitestbase import ApiTestBase


class ScanTestCase(ApiTestBase):
    def setUp(self):
        super().setUp(apiVersion="1.16.0")

    def test_unauthorized(self):
        self._make_request("startScan", args={"u": "bob", "p": "B0b"}, error=50)
        self._make_request("getScanStatus", args={"u": "bob", "p": "B0b"}, error=50)

    def test_unavailable(self):
        self._make_request("startScan", error=0)
        self._make_request("getScanStatus", error=0)


class ScanWithDaemonTestCase(ApiTestBase):
    def setUp(self):
        super().setUp(apiVersion="1.16.0")

        Folder.create(name="Root", root=True, path="tests/assets")

        self._daemon = Daemon(self.config)
        self._thread = Thread(target=self._daemon.run)
        self._thread.start()
        sleep(0.2)  # Wait a bit for the daemon thread to initialize

    def tearDown(self):
        self._daemon.terminate()
        self._thread.join()

        super().tearDown()

    def test_startScan(self):
        rv, child = self._make_request("startScan", tag="scanStatus", skip_post=True)
        self.assertEqual(child.get("scanning"), "true")
        self.assertGreaterEqual(int(child.get("count")), 0)

    def test_getScanStatus(self):
        rv, child = self._make_request("getScanStatus", tag="scanStatus")
        self.assertEqual(child.get("scanning"), "false")
        self.assertEqual(int(child.get("count")), 0)



# This file is part of Supysonic.
# Supysonic is a Python implementation of the Subsonic server API.
#
# Copyright (C) 2013-2024 Alban 'spl0k' Féron
#
# Distributed under terms of the GNU AGPLv3 license.

import importlib
import mimetypes
import os.path
import sys
import time

from datetime import datetime
from hashlib import sha1
from peewee import (
    AutoField,
    BlobField,
    BooleanField,
    CharField,
    DateTimeField,
    FixedCharField,
    ForeignKeyField,
    IntegerField,
    TextField,
    UUIDField,
)
from peewee import CompositeKey, DatabaseProxy, Model, MySQLDatabase
from peewee import fn
from playhouse.db_url import parseresult_to_dict, schemes
from urllib.parse import urlparse
from uuid import UUID, uuid4

SCHEMA_VERSION = "20240318"


def now():
    return datetime.now().replace(microsecond=0)


def random():
    if isinstance(db.obj, MySQLDatabase):
        return fn.rand()
    return fn.random()


def PrimaryKeyField(**kwargs):
    return UUIDField(primary_key=True, default=uuid4, **kwargs)


db = DatabaseProxy()


class _Model(Model):
    class Meta:
        database = db
        legacy_table_names = False


class Meta(_Model):
    key = CharField(32, primary_key=True)
    value = CharField(256)


class PathMixin:
    @classmethod
    def get(cls, *args, **kwargs):
        if kwargs:
            path = kwargs.pop("path", None)
            if path:
                kwargs["_path_hash"] = sha1(path.encode("utf-8")).digest()
        return _Model.get.__func__(cls, *args, **kwargs)

    def __init__(self, *args, **kwargs):
        if "path" in kwargs:
            path = kwargs["path"]
            kwargs["_path_hash"] = sha1(path.encode("utf-8")).digest()
        _Model.__init__(self, *args, **kwargs)

    def __setattr__(self, attr, value):
        _Model.__setattr__(self, attr, value)
        if attr == "path":
            _Model.__setattr__(self, "_path_hash", sha1(value.encode("utf-8")).digest())


class Folder(PathMixin, _Model):
    id = AutoField()
    root = BooleanField()
    name = CharField()
    path = CharField(4096)  # unique
    _path_hash = BlobField(column_name="path_hash", unique=True)
    created = DateTimeField(default=now)
    cover_art = CharField(null=True)
    last_scan = IntegerField(default=0)

    parent = ForeignKeyField("self", null=True, backref="children")

    def as_subsonic_child(self, user):
        info = {
            "id": str(self.id),
            "isDir": True,
            "title": self.name,
            "album": self.name,
            "created": self.created.isoformat(),
        }
        if not self.root:
            info["parent"] = str(self.parent.id)
            info["artist"] = self.parent.name
        if self.cover_art:
            info["coverArt"] = str(self.id)
        else:
            for track in self.tracks:
                if track.has_art:
                    info["coverArt"] = str(track.id)
                    break

        try:
            starred = StarredFolder[user.id, self.id]
            info["starred"] = starred.date.isoformat()
        except StarredFolder.DoesNotExist:
            pass

        try:
            rating = RatingFolder[user.id, self.id]
            info["userRating"] = rating.rating
        except RatingFolder.DoesNotExist:
            pass

        avgRating = (
            RatingFolder.select(fn.avg(RatingFolder.rating, coerce=False))
            .where(RatingFolder.rated == self)
            .scalar()
        )
        if avgRating:
            info["averageRating"] = avgRating

        return info

    def as_subsonic_artist(self, user):  # "Artist" type in XSD
        info = {"id": str(self.id), "name": self.name}

        try:
            starred = StarredFolder[user.id, self.id]
            info["starred"] = starred.date.isoformat()
        except StarredFolder.DoesNotExist:
            pass

        return info

    def as_subsonic_directory(self, user, client):  # "Directory" type in XSD
        info = {
            "id": str(self.id),
            "name": self.name,
            "child": [
                f.as_subsonic_child(user)
                for f in self.children.order_by(fn.lower(Folder.name))
            ]
            + [
                t.as_subsonic_child(user, client)
                for t in sorted(self.tracks, key=lambda t: t.sort_key())
            ],
        }
        if not self.root:
            info["parent"] = str(self.parent.id)

        return info

    @classmethod
    def prune(cls):
        alias = cls.alias()
        query = cls.select(cls.id).where(
            ~cls.root,
            Track.select(fn.count("*")).where(Track.folder == cls.id) == 0,
            alias.select(fn.count("*")).where(alias.parent == cls.id) == 0,
        )
        total = 0
        while True:
            clone = query.clone()  # peewee caches the results, clone to force a refetch
            for f in clone:
                f.delete_instance(recursive=True)
                total += 1
            if not len(clone):
                return total

    def delete_hierarchy(self):
        if self.root:
            cond = Track.root_folder == self
        else:
            cond = Track.path.startswith(self.path)

        return self.__delete_hierarchy(cond)

    def __delete_hierarchy(self, cond):
        users = User.select(User.id).join(Track).where(cond)
        User.update(last_play=None).where(User.id.in_(users)).execute()

        tracks = Track.select(Track.id).where(cond)
        RatingTrack.delete().where(RatingTrack.rated.in_(tracks)).execute()
        StarredTrack.delete().where(StarredTrack.starred.in_(tracks)).execute()

        path_cond = Folder.path.startswith(self.path)
        folders = Folder.select(Folder.id).where(path_cond)
        RatingFolder.delete().where(RatingFolder.rated.in_(folders)).execute()
        StarredFolder.delete().where(StarredFolder.starred.in_(folders)).execute()

        deleted_tracks = Track.delete().where(cond).execute()

        query = Folder.delete().where(path_cond)
        if isinstance(db.obj, MySQLDatabase):
            # MySQL can't propery resolve deletion order when it has several to handle
            query = query.order_by(Folder.path.desc())
        query.execute()

        return deleted_tracks


class Artist(_Model):
    id = PrimaryKeyField()
    name = CharField()

    def as_subsonic_artist(self, user):
        info = {
            "id": str(self.id),
            "name": self.name,
            # coverArt
            "albumCount": self.albums.count(),
        }

        try:
            starred = StarredArtist[user.id, self.id]
            info["starred"] = starred.date.isoformat()
        except StarredArtist.DoesNotExist:
            pass

        return info

    @classmethod
    def prune(cls):
        album_artists = Album.select(Album.artist)
        track_artists = Track.select(Track.artist)

        StarredArtist.delete().where(
            StarredArtist.starred.not_in(album_artists),
            StarredArtist.starred.not_in(track_artists),
        ).execute()

        return (
            cls.delete()
            .where(
                cls.id.not_in(album_artists),
                cls.id.not_in(track_artists),
            )
            .execute()
        )


class Album(_Model):
    id = PrimaryKeyField()
    name = CharField()
    artist = ForeignKeyField(Artist, backref="albums")

    def as_subsonic_album(self, user):  # "AlbumID3" type in XSD
        duration, created, year = self.tracks.select(
            fn.sum(Track.duration), fn.min(Track.created), fn.min(Track.year)
        ).scalar(as_tuple=True)

        info = {
            "id": str(self.id),
            "name": self.name,
            "artist": self.artist.name,
            "artistId": str(self.artist.id),
            "songCount": self.tracks.count(),
            "duration": duration,
            "created": created.isoformat(),
        }

        track_with_cover = (
            self.tracks.join(Folder).where(Folder.cover_art.is_null(False)).first()
        )
        if track_with_cover is not None:
            info["coverArt"] = str(track_with_cover.folder.id)
        else:
            track_with_cover = self.tracks.where(Track.has_art).first()
            if track_with_cover is not None:
                info["coverArt"] = str(track_with_cover.id)

        if year:
            info["year"] = year

        genre = ", ".join(
            g
            for (g,) in self.tracks.select(Track.genre)
            .where(Track.genre.is_null(False))
            .distinct()
            .tuples()
        )
        if genre:
            info["genre"] = genre

        try:
            starred = StarredAlbum[user.id, self.id]
            info["starred"] = starred.date.isoformat()
        except StarredAlbum.DoesNotExist:
            pass

        return info

    def sort_key(self):
        year = self.tracks.select(fn.min(Track.year)).scalar() or 9999
        return f"{year}{self.name.lower()}"

    @classmethod
    def prune(cls):
        albums = Track.select(Track.album)
        StarredAlbum.delete().where(StarredAlbum.starred.not_in(albums)).execute()
        return cls.delete().where(cls.id.not_in(albums)).execute()


class Track(PathMixin, _Model):
    id = PrimaryKeyField()
    disc = IntegerField()
    number = IntegerField()
    title = CharField()
    year = IntegerField(null=True)
    genre = CharField(null=True)
    duration = IntegerField()
    has_art = BooleanField(default=False)

    album = ForeignKeyField(Album, backref="tracks")
    artist = ForeignKeyField(Artist, backref="tracks")

    bitrate = IntegerField()

    path = CharField(4096)  # unique
    _path_hash = BlobField(column_name="path_hash", unique=True)
    created = DateTimeField(default=now)
    last_modification = IntegerField()

    play_count = IntegerField(default=0)
    last_play = DateTimeField(null=True)

    root_folder = ForeignKeyField(Folder, backref="+")
    folder = ForeignKeyField(Folder, backref="tracks")

    def as_subsonic_child(self, user, prefs):
        info = {
            "id": str(self.id),
            "parent": str(self.folder.id),
            "isDir": False,
            "title": self.title,
            "album": self.album.name,
            "artist": self.artist.name,
            "track": self.number,
            "size": os.path.getsize(self.path) if os.path.isfile(self.path) else -1,
            "contentType": self.mimetype,
            "suffix": self.suffix(),
            "duration": self.duration,
            "bitRate": self.bitrate,
            "path": self.path[len(self.root_folder.path) + 1 :],
            "isVideo": False,
            "discNumber": self.disc,
            "created": self.created.isoformat(),
            "albumId": str(self.album.id),
            "artistId": str(self.artist.id),
            "type": "music",
        }

        if self.year:
            info["year"] = self.year
        if self.genre:
            info["genre"] = self.genre
        if self.has_art:
            info["coverArt"] = str(self.id)
        elif self.folder.cover_art:
            info["coverArt"] = str(self.folder.id)

        try:
            starred = StarredTrack[user.id, self.id]
            info["starred"] = starred.date.isoformat()
        except StarredTrack.DoesNotExist:
            pass

        try:
            rating = RatingTrack[user.id, self.id]
            info["userRating"] = rating.rating
        except RatingTrack.DoesNotExist:
            pass

        avgRating = (
            RatingTrack.select(fn.avg(RatingTrack.rating, coerce=False))
            .where(RatingTrack.rated == self)
            .scalar()
        )
        if avgRating:
            info["averageRating"] = avgRating

        if (
            prefs is not None
            and prefs.format is not None
            and prefs.format != self.suffix()
        ):
            info["transcodedSuffix"] = prefs.format
            info["transcodedContentType"] = (
                mimetypes.guess_type("dummyname." + prefs.format, False)[0]
                or "application/octet-stream"
            )

        return info

    @property
    def mimetype(self):
        return mimetypes.guess_type(self.path, False)[0] or "application/octet-stream"

    def duration_str(self):
        ret = f"{(self.duration % 3600) / 60:02}:{self.duration % 60:02}"
        if self.duration >= 3600:
            ret = f"{self.duration / 3600:02}:{ret}"
        return ret

    def suffix(self):
        return os.path.splitext(self.path)[1][1:].lower()

    def sort_key(self):
        return f"{self.album.artist.name}{self.album.name}{self.disc:02}{self.number:02}{self.title}".lower()


class User(_Model):
    id = PrimaryKeyField()
    name = CharField(64, unique=True)
    mail = CharField(null=True)
    password = FixedCharField(40)
    salt = FixedCharField(6)

    admin = BooleanField(default=False)
    jukebox = BooleanField(default=False)

    lastfm_session = FixedCharField(32, null=True)
    lastfm_status = BooleanField(
        default=True
    )  # True: ok/unlinked, False: invalid session

    listenbrainz_session = FixedCharField(36, null=True)
    listenbrainz_status = BooleanField(
        default=True
    )  # True: ok/unlinked, False: invalid token

    last_play = ForeignKeyField(Track, null=True, backref="+")
    last_play_date = DateTimeField(null=True)

    def as_subsonic_user(self):
        return {
            "username": self.name,
            "email": self.mail or "",
            "scrobblingEnabled": self.lastfm_session is not None and self.lastfm_status,
            "adminRole": self.admin,
            "settingsRole": True,
            "downloadRole": True,
            "uploadRole": False,
            "playlistRole": True,
            "coverArtRole": False,
            "commentRole": False,
            "podcastRole": False,
            "streamRole": True,
            "jukeboxRole": self.admin or self.jukebox,
            "shareRole": False,
        }


class ClientPrefs(_Model):
    user = ForeignKeyField(User, backref="clients")
    client_name = CharField(32)
    format = CharField(8, null=True)
    bitrate = IntegerField(null=True)

    class Meta:
        primary_key = CompositeKey("user", "client_name")


def _make_starred_model(target_model):
    class Starred(_Model):
        user = ForeignKeyField(User, backref="+")
        starred = ForeignKeyField(target_model, backref="+")
        date = DateTimeField(default=now)

        class Meta:
            primary_key = CompositeKey("user", "starred")
            table_name = "starred_" + target_model._meta.table_name

    return Starred


StarredFolder = _make_starred_model(Folder)
StarredArtist = _make_starred_model(Artist)
StarredAlbum = _make_starred_model(Album)
StarredTrack = _make_starred_model(Track)


def _make_rating_model(target_model):
    class Rating(_Model):
        user = ForeignKeyField(User, backref="+")
        rated = ForeignKeyField(target_model, backref="+")
        rating = IntegerField()  # min=1, max=5

        class Meta:
            primary_key = CompositeKey("user", "rated")
            table_name = "rating_" + target_model._meta.table_name

    return Rating


RatingFolder = _make_rating_model(Folder)
RatingTrack = _make_rating_model(Track)


class ChatMessage(_Model):
    id = PrimaryKeyField()
    user = ForeignKeyField(User, backref="+")
    time = IntegerField(default=lambda: int(time.time()))
    message = CharField(512)

    def responsize(self):
        return {
            "username": self.user.name,
            "time": self.time * 1000,
            "message": self.message,
        }


class Playlist(_Model):
    id = PrimaryKeyField()
    user = ForeignKeyField(User, backref="playlists")
    name = CharField()
    comment = CharField(null=True)
    public = BooleanField(default=False)
    created = DateTimeField(default=now)
    tracks = TextField(null=True)

    def as_subsonic_playlist(self, user):
        tracks = self.get_tracks()
        info = {
            "id": str(self.id),
            "name": (
                self.name
                if self.user.id == user.id
                else f"[{self.user.name}] {self.name}"
            ),
            "owner": self.user.name,
            "public": self.public,
            "songCount": len(tracks),
            "duration": sum(t.duration for t in tracks),
            "created": self.created.isoformat(),
        }
        if self.comment:
            info["comment"] = self.comment
        return info

    def get_tracks(self):
        if not self.tracks:
            return []

        tracks = []
        should_fix = False

        for t in self.tracks.split(","):
            try:
                tid = UUID(t)
                track = Track[tid]
                tracks.append(track)
            except (ValueError, Track.DoesNotExist):
                should_fix = True

        if should_fix:
            self.tracks = ",".join(str(t.id) for t in tracks)
            db.commit()

        return tracks

    def clear(self):
        self.tracks = ""

    def add(self, track):
        if isinstance(track, UUID):
            tid = track
        elif isinstance(track, Track):
            tid = track.id
        elif isinstance(track, str):
            tid = UUID(track)

        if self.tracks and len(self.tracks) > 0:
            self.tracks = f"{self.tracks},{tid}"
        else:
            self.tracks = str(tid)

    def remove_at_indexes(self, indexes):
        tracks = self.tracks.split(",")
        for i in indexes:
            if i < 0 or i >= len(tracks):
                continue
            tracks[i] = None

        self.tracks = ",".join(t for t in tracks if t)


class RadioStation(_Model):
    id = PrimaryKeyField()
    stream_url = CharField()
    name = CharField()
    homepage_url = CharField(null=True)
    created = DateTimeField(default=now)

    def as_subsonic_station(self):
        info = {
            "id": str(self.id),
            "streamUrl": self.stream_url,
            "name": self.name,
            "homePageUrl": self.homepage_url,
        }
        return info


if sys.version_info < (3, 9):
    import pkg_resources

    def get_resource_text(respath):
        return pkg_resources.resource_string(__package__, respath).decode("utf-8")

    def list_migrations(provider):
        return pkg_resources.resource_listdir(
            __package__, f"schema/migration/{provider}"
        )

else:
    import importlib.resources

    def get_resource_text(respath):
        return (
            importlib.resources.files(__package__).joinpath(respath).read_text("utf-8")
        )

    def list_migrations(provider):
        return (
            e.name
            for e in importlib.resources.files(__package__)
            .joinpath(f"schema/migration/{provider}")
            .iterdir()
        )


def execute_sql_resource_script(respath):
    sql = get_resource_text(respath)
    for statement in sql.split(";"):
        statement = statement.strip()
        if statement and not statement.startswith("--"):
            db.execute_sql(statement)


def init_database(database_uri):
    uri = urlparse(database_uri)
    args = parseresult_to_dict(uri)
    if uri.scheme.startswith("mysql"):
        args.setdefault("charset", "utf8mb4")
        args.setdefault("binary_prefix", True)

    if uri.scheme.startswith("mysql"):
        provider = "mysql"
    elif uri.scheme.startswith("postgres"):
        provider = "postgres"
    elif uri.scheme.startswith("sqlite"):
        provider = "sqlite"
        args["pragmas"] = {"foreign_keys": 1}
    else:
        raise RuntimeError(f"Unsupported database: {uri.scheme}")

    db_class = schemes.get(uri.scheme)
    db.initialize(db_class(**args))
    db.connect()

    # Check if we should create the tables
    if not db.table_exists("meta"):
        with db.atomic():
            execute_sql_resource_script(f"schema/{provider}.sql")
            Meta.create(key="schema_version", value=SCHEMA_VERSION)

    # Check for schema changes
    version = Meta["schema_version"]
    if version.value < SCHEMA_VERSION:
        args.pop("pragmas", ())
        migrations = sorted(list_migrations(provider))
        for migration in migrations:
            if migration[0] in ("_", "."):
                continue

            date, ext = os.path.splitext(migration)
            if date <= version.value:
                continue

            if ext == ".sql":
                with db.atomic():
                    execute_sql_resource_script(
                        f"schema/migration/{provider}/{migration}"
                    )
            elif ext == ".py":
                m = importlib.import_module(
                    f".schema.migration.{provider}.{date}", __package__
                )
                m.apply(args.copy())

        version.value = SCHEMA_VERSION
        version.save()


def release_database():
    db.close()
    db.initialize(None)


def open_connection(reuse=False):
    return db.connect(reuse)


def close_connection():
    db.close()



# This file is part of Supysonic.
# Supysonic is a Python implementation of the Subsonic server API.
#
# Copyright (C) 2017-2022 Alban 'spl0k' Féron
#
# Distributed under terms of the GNU AGPLv3 license.

import os
import tempfile
import shlex
import unittest

from click.testing import CliRunner

from supysonic.db import Folder, User, init_database, release_database
from supysonic.cli import cli

from ..testbase import TestConfig


class CLITestCase(unittest.TestCase):
    """Really basic tests. Some even don't check anything but are just there for coverage"""

    def setUp(self):
        self.__conf = TestConfig(False, False)
        self.__db = tempfile.mkstemp()
        self.__conf.BASE["database_uri"] = "sqlite:///" + self.__db[1]
        init_database(self.__conf.BASE["database_uri"])

        self.__runner = CliRunner()

    def tearDown(self):
        release_database()
        os.close(self.__db[0])
        os.remove(self.__db[1])

    def __invoke(self, cmd, expect_fail=False):
        rv = self.__runner.invoke(cli, shlex.split(cmd), obj=self.__conf)
        func = self.assertNotEqual if expect_fail else self.assertEqual
        func(rv.exit_code, 0)
        return rv

    def __add_folder(self, name, path, expect_fail=False):
        self.__invoke(f"folder add {name} {shlex.quote(path)}", expect_fail)

    def test_folder_add(self):
        with tempfile.TemporaryDirectory() as d:
            self.__add_folder("tmpfolder", d)

        f = Folder.select().first()
        self.assertIsNotNone(f)
        self.assertEqual(f.path, d)

    def test_folder_add_errors(self):
        with tempfile.TemporaryDirectory() as d:
            self.__add_folder("f1", d)
            self.__add_folder("f2", d, True)
        with tempfile.TemporaryDirectory() as d:
            self.__add_folder("f1", d, True)
        self.__invoke("folder add f3 /invalid/path", True)

        self.assertEqual(Folder.select().count(), 1)

    def test_folder_delete(self):
        with tempfile.TemporaryDirectory() as d:
            self.__add_folder("tmpfolder", d)
        self.__invoke("folder delete randomfolder", True)
        self.__invoke("folder delete tmpfolder")

        self.assertEqual(Folder.select().count(), 0)

    def test_folder_list(self):
        with tempfile.TemporaryDirectory() as d:
            self.__add_folder("tmpfolder", d)
            rv = self.__invoke("folder list")
            self.assertIn("tmpfolder", rv.output)
            self.assertIn(d, rv.output)

    def test_folder_scan(self):
        with tempfile.TemporaryDirectory() as d:
            self.__add_folder("tmpfolder", d)
            with tempfile.NamedTemporaryFile(dir=d):
                self.__invoke("folder scan")
                self.__invoke("folder scan tmpfolder nonexistent")

    def test_user_add(self):
        self.__invoke("user add -p Alic3 alice")
        self.__invoke("user add -p alice alice", True)

        self.assertEqual(User.select().count(), 1)

    def test_user_delete(self):
        self.__invoke("user add -p Alic3 alice")
        self.__invoke("user delete alice")
        self.__invoke("user delete bob", True)

        self.assertEqual(User.select().count(), 0)

    def test_user_list(self):
        self.__invoke("user add -p Alic3 alice")
        rv = self.__invoke("user list")
        self.assertIn("alice", rv.output)

    def test_user_setadmin(self):
        self.__invoke("user add -p Alic3 alice")
        self.__invoke("user setroles -A alice")
        self.__invoke("user setroles -A bob", True)
        self.assertTrue(User.get(name="alice").admin)

    def test_user_unsetadmin(self):
        self.__invoke("user add -p Alic3 alice")
        self.__invoke("user setroles -A alice")
        self.__invoke("user setroles -a alice")
        self.assertFalse(User.get(name="alice").admin)

    def test_user_setjukebox(self):
        self.__invoke("user add -p Alic3 alice")
        self.__invoke("user setroles -J alice")
        self.assertTrue(User.get(name="alice").jukebox)

    def test_user_unsetjukebox(self):
        self.__invoke("user add -p Alic3 alice")
        self.__invoke("user setroles -J alice")
        self.__invoke("user setroles -j alice")
        self.assertFalse(User.get(name="alice").jukebox)

    def test_user_changepass(self):
        self.__invoke("user add -p Alic3 alice")
        self.__invoke("user changepass alice -p newpass")
        self.__invoke("user changepass bob -p B0b", True)

    def test_user_rename(self):
        self.__invoke("user add -p Alic3 alice")
        self.__invoke("user rename alice alice")
        self.__invoke("user rename bob charles", True)

        self.__invoke("user rename alice ''", True)
        self.assertEqual(User.select().first().name, "alice")

        self.__invoke("user rename alice bob")
        self.assertEqual(User.select().first().name, "bob")

        self.__invoke("user add -p Ch4rl3s charles")
        self.__invoke("user rename bob charles", True)
        self.assertEqual(User.select().where(User.name == "bob").count(), 1)
        self.assertEqual(User.select().where(User.name == "charles").count(), 1)


if __name__ == "__main__":
    unittest.main()



# This file is part of Supysonic.
# Supysonic is a Python implementation of the Subsonic server API.
#
# Copyright (C) 2017-2022 Alban 'spl0k' Féron
#
# Distributed under terms of the GNU AGPLv3 license.

import unittest
import uuid

from supysonic.db import Folder, Artist, Album, Track, Playlist, User

from .apitestbase import ApiTestBase


class PlaylistTestCase(ApiTestBase):
    def setUp(self):
        super().setUp()

        root = Folder.create(root=True, name="Root folder", path="tests/assets")
        artist = Artist.create(name="Artist")
        album = Album.create(name="Album", artist=artist)

        songs = {}
        for num, song in enumerate(["One", "Two", "Three", "Four"]):
            track = Track.create(
                disc=1,
                number=num,
                title=song,
                duration=2,
                album=album,
                artist=artist,
                bitrate=320,
                path="tests/assets/" + song,
                last_modification=0,
                root_folder=root,
                folder=root,
            )
            songs[song] = track

        users = {u.name: u for u in User.select()}

        playlist = Playlist.create(user=users["alice"], name="Alice's")
        playlist.add(songs["One"])
        playlist.add(songs["Three"])
        playlist.save()

        playlist = Playlist.create(
            user=users["alice"], public=True, name="Alice's public"
        )
        playlist.add(songs["One"])
        playlist.add(songs["Two"])
        playlist.save()

        playlist = Playlist.create(user=users["bob"], name="Bob's")
        playlist.add(songs["Two"])
        playlist.add(songs["Four"])
        playlist.save()

    def test_get_playlists(self):
        # get own playlists
        rv, child = self._make_request("getPlaylists", tag="playlists")
        self.assertEqual(len(child), 2)
        self.assertEqual(child[0].get("owner"), "alice")
        self.assertEqual(child[1].get("owner"), "alice")

        # get own and public
        rv, child = self._make_request(
            "getPlaylists", {"u": "bob", "p": "B0b"}, tag="playlists"
        )
        self.assertEqual(len(child), 2)
        self.assertTrue("alice" in (child[0].get("owner"), child[1].get("owner")))
        self.assertTrue("bob" in (child[0].get("owner"), child[1].get("owner")))
        self.assertIsNotNone(
            self._find(child, "./playlist[@owner='alice'][@public='true']")
        )

        # get other
        rv, child = self._make_request(
            "getPlaylists", {"username": "bob"}, tag="playlists"
        )
        self.assertEqual(len(child), 1)
        self.assertEqual(child[0].get("owner"), "bob")

        # get other when not admin
        self._make_request(
            "getPlaylists", {"u": "bob", "p": "B0b", "username": "alice"}, error=50
        )

        # get from unknown user
        self._make_request("getPlaylists", {"username": "johndoe"}, error=70)

    def test_get_playlist(self):
        # missing param
        self._make_request("getPlaylist", error=10)

        # invalid id
        self._make_request("getPlaylist", {"id": 1234}, error=0)

        # unknown
        self._make_request("getPlaylist", {"id": str(uuid.uuid4())}, error=70)

        # other's private from non admin
        playlist = (
            Playlist.select()
            .join(User)
            .where(~Playlist.public, User.name == "alice")
            .get()
        )
        self._make_request(
            "getPlaylist", {"u": "bob", "p": "B0b", "id": str(playlist.id)}, error=50
        )

        # standard
        rv, child = self._make_request("getPlaylists", tag="playlists")
        self._make_request("getPlaylist", {"id": child[0].get("id")}, tag="playlist")
        rv, child = self._make_request(
            "getPlaylist", {"id": child[1].get("id")}, tag="playlist"
        )
        self.assertEqual(child.get("songCount"), "2")
        self.assertEqual(
            self._xpath(child, "count(./entry)"), 2
        )  # don't count children, there may be 'allowedUser's (even though not supported by supysonic)
        self.assertEqual(child.get("duration"), "4")
        self.assertEqual(child[0].get("title"), "One")
        self.assertTrue(
            child[1].get("title") in ("Two", "Three")
        )  # depending on 'getPlaylists' result ordering

    def test_create_playlist(self):
        self._make_request("createPlaylist", error=10)
        self._make_request(
            "createPlaylist", {"name": "wrongId", "songId": "abc"}, error=0
        )
        self._make_request(
            "createPlaylist",
            {"name": "unknownId", "songId": str(uuid.uuid4())},
            error=70,
        )
        self._make_request("createPlaylist", {"playlistId": "abc"}, error=0)
        self._make_request(
            "createPlaylist", {"playlistId": str(uuid.uuid4())}, error=70
        )

        # create
        self._make_request("createPlaylist", {"name": "new playlist"}, skip_post=True)
        rv, child = self._make_request("getPlaylists", tag="playlists")
        self.assertEqual(len(child), 3)
        playlist = self._find(child, "./playlist[@name='new playlist']")
        self.assertEqual(len(playlist), 0)

        # "update" newly created
        self._make_request("createPlaylist", {"playlistId": playlist.get("id")})
        rv, child = self._make_request("getPlaylists", tag="playlists")
        self.assertEqual(len(child), 3)

        # renaming
        self._make_request(
            "createPlaylist", {"playlistId": playlist.get("id"), "name": "renamed"}
        )
        rv, child = self._make_request("getPlaylists", tag="playlists")
        self.assertEqual(len(child), 3)
        self.assertIsNone(self._find(child, "./playlist[@name='new playlist']"))
        playlist = self._find(child, "./playlist[@name='renamed']")
        self.assertIsNotNone(playlist)

        # update from other user
        self._make_request(
            "createPlaylist",
            {"u": "bob", "p": "B0b", "playlistId": playlist.get("id")},
            error=50,
        )

        # create more useful playlist
        songs = {s.title: str(s.id) for s in Track.select()}
        self._make_request(
            "createPlaylist",
            {
                "name": "songs",
                "songId": [songs[s] for s in ("Three", "One", "Two")],
            },
            skip_post=True,
        )
        playlist = Playlist.get(name="songs")
        self.assertIsNotNone(playlist)
        rv, child = self._make_request(
            "getPlaylist", {"id": str(playlist.id)}, tag="playlist"
        )
        self.assertEqual(child.get("songCount"), "3")
        self.assertEqual(self._xpath(child, "count(./entry)"), 3)
        self.assertEqual(child[0].get("title"), "Three")
        self.assertEqual(child[1].get("title"), "One")
        self.assertEqual(child[2].get("title"), "Two")

        # update
        self._make_request(
            "createPlaylist",
            {"playlistId": str(playlist.id), "songId": songs["Two"]},
            skip_post=True,
        )
        rv, child = self._make_request(
            "getPlaylist", {"id": str(playlist.id)}, tag="playlist"
        )
        self.assertEqual(child.get("songCount"), "1")
        self.assertEqual(self._xpath(child, "count(./entry)"), 1)
        self.assertEqual(child[0].get("title"), "Two")

    def assertPlaylistCountEqual(self, count):
        self.assertEqual(Playlist.select().count(), count)

    def test_delete_playlist(self):
        # check params
        self._make_request("deletePlaylist", error=10)
        self._make_request("deletePlaylist", {"id": "string"}, error=0)
        self._make_request("deletePlaylist", {"id": str(uuid.uuid4())}, error=70)

        # delete unowned when not admin
        playlist = Playlist.select().join(User).where(User.name == "alice").first()
        self._make_request(
            "deletePlaylist", {"u": "bob", "p": "B0b", "id": str(playlist.id)}, error=50
        )
        self.assertPlaylistCountEqual(3)

        # delete owned
        self._make_request("deletePlaylist", {"id": str(playlist.id)}, skip_post=True)
        self.assertPlaylistCountEqual(2)
        self._make_request("deletePlaylist", {"id": str(playlist.id)}, error=70)
        self.assertPlaylistCountEqual(2)

        # delete unowned when admin
        playlist = Playlist.select().join(User).where(User.name == "bob").get()
        self._make_request("deletePlaylist", {"id": str(playlist.id)}, skip_post=True)
        self.assertPlaylistCountEqual(1)

    def test_update_playlist(self):
        self._make_request("updatePlaylist", error=10)
        self._make_request("updatePlaylist", {"playlistId": 1234}, error=0)
        self._make_request(
            "updatePlaylist", {"playlistId": str(uuid.uuid4())}, error=70
        )

        playlist = (
            Playlist.select()
            .join(User)
            .where(User.name == "alice")
            .order_by(Playlist.created)
            .first()
        )
        pid = str(playlist.id)
        self._make_request(
            "updatePlaylist", {"playlistId": pid, "songIdToAdd": "string"}, error=0
        )
        self._make_request(
            "updatePlaylist",
            {"playlistId": pid, "songIndexToRemove": "string"},
            error=0,
        )

        name = str(playlist.name)
        self._make_request(
            "updatePlaylist",
            {"u": "bob", "p": "B0b", "playlistId": pid, "name": "new name"},
            error=50,
        )
        rv, child = self._make_request("getPlaylist", {"id": pid}, tag="playlist")
        self.assertEqual(child.get("name"), name)
        self.assertEqual(self._xpath(child, "count(./entry)"), 2)

        self._make_request(
            "updatePlaylist", {"playlistId": pid, "name": "new name"}, skip_post=True
        )
        rv, child = self._make_request("getPlaylist", {"id": pid}, tag="playlist")
        self.assertEqual(child.get("name"), "new name")
        self.assertEqual(self._xpath(child, "count(./entry)"), 2)

        self._make_request(
            "updatePlaylist",
            {"playlistId": pid, "songIndexToRemove": [-1, 2]},
            skip_post=True,
        )
        rv, child = self._make_request("getPlaylist", {"id": pid}, tag="playlist")
        self.assertEqual(self._xpath(child, "count(./entry)"), 2)

        self._make_request(
            "updatePlaylist",
            {"playlistId": pid, "songIndexToRemove": [0, 2]},
            skip_post=True,
        )
        rv, child = self._make_request("getPlaylist", {"id": pid}, tag="playlist")
        self.assertEqual(self._xpath(child, "count(./entry)"), 1)
        self.assertEqual(self._find(child, "./entry").get("title"), "Three")

        songs = {s.title: str(s.id) for s in Track.select()}

        self._make_request(
            "updatePlaylist",
            {
                "playlistId": pid,
                "songIdToAdd": [songs["One"], songs["Two"], songs["Two"]],
            },
            skip_post=True,
        )
        rv, child = self._make_request("getPlaylist", {"id": pid}, tag="playlist")
        self.assertSequenceEqual(
            self._xpath(child, "./entry/@title"), ["Three", "One", "Two", "Two"]
        )

        self._make_request(
            "updatePlaylist",
            {"playlistId": pid, "songIndexToRemove": [2, 1]},
            skip_post=True,
        )
        rv, child = self._make_request("getPlaylist", {"id": pid}, tag="playlist")
        self.assertSequenceEqual(self._xpath(child, "./entry/@title"), ["Three", "Two"])

        self._make_request(
            "updatePlaylist",
            {"playlistId": pid, "songIdToAdd": songs["One"]},
            skip_post=True,
        )
        self._make_request(
            "updatePlaylist",
            {"playlistId": pid, "songIndexToRemove": [1, 1]},
            skip_post=True,
        )
        rv, child = self._make_request("getPlaylist", {"id": pid}, tag="playlist")
        self.assertSequenceEqual(self._xpath(child, "./entry/@title"), ["Three", "One"])

        self._make_request(
            "updatePlaylist",
            {"playlistId": pid, "songIdToAdd": str(uuid.uuid4())},
            error=70,
        )
        rv, child = self._make_request("getPlaylist", {"id": pid}, tag="playlist")
        self.assertEqual(self._xpath(child, "count(./entry)"), 2)


if __name__ == "__main__":
    unittest.main()



# This file is part of Supysonic.
# Supysonic is a Python implementation of the Subsonic server API.
#
# Copyright (C) 2020-2022 Alban 'spl0k' Féron
#
# Distributed under terms of the GNU AGPLv3 license.

from flask import request

from ..db import RadioStation

from . import get_entity, api_routing
from .exceptions import Forbidden, MissingParameter


@api_routing("/getInternetRadioStations")
def get_radio_stations():
    query = RadioStation.select().order_by(RadioStation.name)
    return request.formatter(
        "internetRadioStations",
        {"internetRadioStation": [p.as_subsonic_station() for p in query]},
    )


@api_routing("/createInternetRadioStation")
def create_radio_station():
    if not request.user.admin:
        raise Forbidden()

    stream_url, name, homepage_url = map(
        request.values.get, ("streamUrl", "name", "homepageUrl")
    )

    if stream_url and name:
        RadioStation.create(stream_url=stream_url, name=name, homepage_url=homepage_url)
    else:
        raise MissingParameter("streamUrl or name")

    return request.formatter.empty


@api_routing("/updateInternetRadioStation")
def update_radio_station():
    if not request.user.admin:
        raise Forbidden()

    res = get_entity(RadioStation)

    stream_url, name, homepage_url = map(
        request.values.get, ("streamUrl", "name", "homepageUrl")
    )
    if stream_url and name:
        res.stream_url = stream_url
        res.name = name

        if homepage_url:
            res.homepage_url = homepage_url

        res.save()
    else:
        raise MissingParameter("streamUrl or name")

    return request.formatter.empty


@api_routing("/deleteInternetRadioStation")
def delete_radio_station():
    if not request.user.admin:
        raise Forbidden()

    res = get_entity(RadioStation)
    res.delete_instance()

    return request.formatter.empty



# This file is part of Supysonic.
# Supysonic is a Python implementation of the Subsonic server API.
#
# Copyright (C) 2017 Alban 'spl0k' Féron
#
# Distributed under terms of the GNU AGPLv3 license.

import unittest

import time

from .apitestbase import ApiTestBase


class ChatTestCase(ApiTestBase):
    def test_add_message(self):
        self._make_request("addChatMessage", error=10)
        rv, child = self._make_request("getChatMessages", tag="chatMessages")
        self.assertEqual(len(child), 0)

        self._make_request(
            "addChatMessage", {"message": "Heres a message"}, skip_post=True
        )
        rv, child = self._make_request("getChatMessages", tag="chatMessages")
        self.assertEqual(len(child), 1)
        self.assertEqual(child[0].get("username"), "alice")
        self.assertEqual(child[0].get("message"), "Heres a message")

    def test_get_messages(self):
        self._make_request("addChatMessage", {"message": "Hello"}, skip_post=True)
        time.sleep(1)
        self._make_request(
            "addChatMessage", {"message": "Is someone there?"}, skip_post=True
        )

        rv, child = self._make_request("getChatMessages", tag="chatMessages")
        self.assertEqual(len(child), 2)

        rv, child = self._make_request(
            "getChatMessages",
            {"since": int(time.time()) * 1000 - 500},
            tag="chatMessages",
        )
        self.assertEqual(len(child), 1)
        self.assertEqual(child[0].get("message"), "Is someone there?")

        self._make_request("getChatMessages", {"since": "invalid timestamp"}, error=0)


if __name__ == "__main__":
    unittest.main()



import argparse
import hashlib
import sqlite3

try:
    bytes = buffer
except NameError:
    pass

parser = argparse.ArgumentParser()
parser.add_argument("dbfile", help="Path to the SQLite database file")
args = parser.parse_args()


def process_table(connection, table):
    c = connection.cursor()

    c.execute(
        "ALTER TABLE {} ADD COLUMN path_hash BLOB NOT NULL DEFAULT ROWID".format(table)
    )

    hashes = dict()
    for row in c.execute("SELECT path FROM {}".format(table)):
        hashes[row[0]] = hashlib.sha1(row[0].encode("utf-8")).digest()
    c.executemany(
        "UPDATE {} SET path_hash=? WHERE path=?".format(table),
        [(bytes(h), p) for p, h in hashes.items()],
    )

    c.execute("CREATE UNIQUE INDEX index_{0}_path ON {0}(path_hash)".format(table))


with sqlite3.connect(args.dbfile) as conn:
    process_table(conn, "folder")
    process_table(conn, "track")
    conn.cursor().execute("VACUUM")



# This file is part of Supysonic.
# Supysonic is a Python implementation of the Subsonic server API.
#
# Copyright (C) 2013-2023 Alban 'spl0k' Féron
#
# Distributed under terms of the GNU AGPLv3 license.

import logging

from flask import flash, redirect, render_template, request, session, url_for
from flask import current_app
from functools import wraps

from ..db import ClientPrefs, User
from ..lastfm import LastFm
from ..listenbrainz import ListenBrainz
from ..managers.user import UserManager

from . import admin_only, frontend

logger = logging.getLogger(__name__)


def me_or_uuid(f, arg="uid"):
    @wraps(f)
    def decorated_func(*args, **kwargs):
        if kwargs:
            uid = kwargs[arg]
        else:
            uid = args[0]

        if uid == "me":
            user = request.user
        elif not request.user.admin:
            return redirect(url_for("frontend.index"))
        else:
            try:
                user = UserManager.get(uid)
            except ValueError as e:
                flash(str(e), "error")
                return redirect(url_for("frontend.index"))
            except User.DoesNotExist:
                flash("No such user", "error")
                return redirect(url_for("frontend.index"))

        if kwargs:
            kwargs["user"] = user
        else:
            args = (uid, user)

        return f(*args, **kwargs)

    return decorated_func


@frontend.route("/user")
@admin_only
def user_index():
    return render_template("users.html", users=User.select())


@frontend.route("/user/<uid>")
@me_or_uuid
def user_profile(uid, user):
    return render_template(
        "profile.html",
        user=user,
        api_key=current_app.config["LASTFM"]["api_key"],
        clients=user.clients,
    )


@frontend.route("/user/<uid>", methods=["POST"])
@me_or_uuid
def update_clients(uid, user):
    clients_opts = {}
    for key, value in request.form.items():
        if "_" not in key:
            continue
        parts = key.split("_")
        if len(parts) != 2:
            continue
        client, opt = parts
        if not client or not opt:
            continue

        if client not in clients_opts:
            clients_opts[client] = {opt: value}
        else:
            clients_opts[client][opt] = value
    logger.debug(clients_opts)

    for client, opts in clients_opts.items():
        prefs = user.clients.where(ClientPrefs.client_name == client).first()
        if prefs is None:
            continue

        if "delete" in opts and opts["delete"] in [
            "on",
            "true",
            "checked",
            "selected",
            "1",
        ]:
            prefs.delete_instance()
            continue

        prefs.format = opts["format"] if "format" in opts and opts["format"] else None
        prefs.bitrate = (
            int(opts["bitrate"]) if "bitrate" in opts and opts["bitrate"] else None
        )
        prefs.save()

    flash("Clients preferences updated.")
    return user_profile(uid, user)


@frontend.route("/user/<uid>/changeusername")
@admin_only
def change_username_form(uid):
    try:
        user = UserManager.get(uid)
    except ValueError as e:
        flash(str(e), "error")
        return redirect(url_for("frontend.index"))
    except User.DoesNotExist:
        flash("No such user", "error")
        return redirect(url_for("frontend.index"))

    return render_template("change_username.html", user=user)


@frontend.route("/user/<uid>/changeusername", methods=["POST"])
@admin_only
def change_username_post(uid):
    try:
        user = UserManager.get(uid)
    except ValueError as e:
        flash(str(e), "error")
        return redirect(url_for("frontend.index"))
    except User.DoesNotExist:
        flash("No such user", "error")
        return redirect(url_for("frontend.index"))

    username = request.form.get("user")
    if username in ("", None):
        flash("The username is required")
        return render_template("change_username.html", user=user)
    if user.name != username:
        try:
            User.get(name=username)
            flash("This name is already taken")
            return render_template("change_username.html", user=user)
        except User.DoesNotExist:
            pass

    if request.form.get("admin") is None:
        admin = False
    else:
        admin = True

    if user.name != username or user.admin != admin:
        user.name = username
        user.admin = admin
        user.save()
        flash(f"User '{username}' updated.")
    else:
        flash(f"No changes for '{username}'.")

    return redirect(url_for("frontend.user_profile", uid=uid))


@frontend.route("/user/<uid>/changemail")
@me_or_uuid
def change_mail_form(uid, user):
    return render_template("change_mail.html", user=user)


@frontend.route("/user/<uid>/changemail", methods=["POST"])
@me_or_uuid
def change_mail_post(uid, user):
    mail = request.form.get("mail", "")
    # No validation, lol.
    user.mail = mail
    return redirect(url_for("frontend.user_profile", uid=uid))


@frontend.route("/user/<uid>/changepass")
@me_or_uuid
def change_password_form(uid, user):
    return render_template("change_pass.html", user=user)


@frontend.route("/user/<uid>/changepass", methods=["POST"])
@me_or_uuid
def change_password_post(uid, user):
    error = False
    if user.id == request.user.id:
        current = request.form.get("current")
        if not current:
            flash("The current password is required")
            error = True

    new, confirm = map(request.form.get, ("new", "confirm"))

    if not new:
        flash("The new password is required")
        error = True
    if new != confirm:
        flash("The new password and its confirmation don't match")
        error = True

    if not error:
        try:
            if user.id == request.user.id:
                UserManager.change_password(user.id, current, new)
            else:
                UserManager.change_password2(user.name, new)

            flash("Password changed")
            return redirect(url_for("frontend.user_profile", uid=uid))
        except ValueError as e:
            flash(str(e), "error")

    return change_password_form(uid, user)


@frontend.route("/user/add")
@admin_only
def add_user_form():
    return render_template("adduser.html")


@frontend.route("/user/add", methods=["POST"])
@admin_only
def add_user_post():
    error = False
    args = request.form.copy()
    (name, passwd, passwd_confirm) = map(
        args.pop, ("user", "passwd", "passwd_confirm"), (None,) * 3
    )
    if not name:
        flash("The name is required.")
        error = True
    if not passwd:
        flash("Please provide a password.")
        error = True
    elif passwd != passwd_confirm:
        flash("The passwords don't match.")
        error = True

    if not error:
        try:
            UserManager.add(name, passwd, **args)
            flash(f"User '{name}' successfully added")
            return redirect(url_for("frontend.user_index"))
        except ValueError as e:
            flash(str(e), "error")

    return add_user_form()


@frontend.route("/user/del/<uid>")
@admin_only
def del_user(uid):
    try:
        UserManager.delete(uid)
        flash("Deleted user")
    except ValueError as e:
        flash(str(e), "error")
    except User.DoesNotExist:
        flash("No such user", "error")

    return redirect(url_for("frontend.user_index"))


@frontend.route("/user/<uid>/lastfm/link")
@me_or_uuid
def lastfm_reg(uid, user):
    token = request.args.get("token")
    if not token:
        flash("Missing LastFM auth token")
        return redirect(url_for("frontend.user_profile", uid=uid))

    lfm = LastFm(current_app.config["LASTFM"], user)
    status, error = lfm.link_account(token)
    flash(error if not status else "Successfully linked LastFM account")

    return redirect(url_for("frontend.user_profile", uid=uid))


@frontend.route("/user/<uid>/lastfm/unlink")
@me_or_uuid
def lastfm_unreg(uid, user):
    lfm = LastFm(current_app.config["LASTFM"], user)
    lfm.unlink_account()
    flash("Unlinked LastFM account")
    return redirect(url_for("frontend.user_profile", uid=uid))


@frontend.route("/user/<uid>/listenbrainz/link")
@me_or_uuid
def listenbrainz_reg(uid, user):
    token = request.args.get("token")
    if not token:
        flash("Missing ListenBrainz auth token")
        return redirect(url_for("frontend.user_profile", uid=uid))

    lbz = ListenBrainz(current_app.config["LISTENBRAINZ"], user)
    status, error = lbz.link_account(token)
    flash(error if not status else "Successfully linked ListenBrainz account")

    return redirect(url_for("frontend.user_profile", uid=uid))


@frontend.route("/user/<uid>/listenbrainz/unlink")
@me_or_uuid
def listenbrainz_unreg(uid, user):
    lbz = ListenBrainz(current_app.config["LISTENBRAINZ"], user)
    lbz.unlink_account()
    flash("Unlinked ListenBrainz account")
    return redirect(url_for("frontend.user_profile", uid=uid))


@frontend.route("/user/login", methods=["GET", "POST"])
def login():
    return_url = request.args.get("returnUrl") or url_for("frontend.index")
    if request.user:
        flash("Already logged in")
        return redirect(return_url)

    if request.method == "GET":
        return render_template("login.html")

    name, password = map(request.form.get, ("user", "password"))
    error = False
    if not name:
        flash("Missing user name")
        error = True
    if not password:
        flash("Missing password")
        error = True

    if not error:
        user = UserManager.try_auth(name, password)
        if user:
            logger.info("Logged user %s (IP: %s)", name, request.remote_addr)
            session["userid"] = str(user.id)
            flash("Logged in!")
            return redirect(return_url)
        else:
            logger.error(
                "Failed login attempt for user %s (IP: %s)", name, request.remote_addr
            )
            flash("Wrong username or password")

    return render_template("login.html")


@frontend.route("/user/logout")
def logout():
    session.clear()
    flash("Logged out!")
    return redirect(url_for("frontend.login"))
